#!/bin/sh
# Reference: DS-N3101_AQR111-AQR112_AQR111C-AQR112C_v1.1.8.pdf

########################################################
# Definition
########################################################
BOARD_CFG="/etc/factory.d/board.cfg"

if [ -e $BOARD_CFG ]; then
	. $BOARD_CFG
else
	echo "$BOARD_CFG is not exist!";
	exit 1;
fi

AQR_PHY_LED_CFG="/etc/factory.d/aqr_phy_led.cfg"

if [ -e $AQR_PHY_LED_CFG ]; then
	. $AQR_PHY_LED_CFG
else
	echo "$AQR_PHY_LED_CFG is not exist!";
	exit 1;
fi

########################################################
# Debug message
########################################################
DEBUG=2;

DBG() {
	[ $DEBUG -gt 0 ] && echo "=> $@" > /dev/console
}
dbg() {
	[ $DEBUG -gt 1 ] &&	echo "$@" > /dev/console
}


########################################################
# sub-Functions
########################################################
#-----------------------------------------------------
# usage
#-----------------------------------------------------
usage() {
	echo "Usage:"
	echo "aqr_phy_led.sh [phy_addr] [g_led0_state] [g_led1_state] [g_led2_state]"
	echo "e.g.: aqr_phy_led.sh 8 1 0 1"
}

#-----------------------------------------------------
# aqr_phy_led_set
# $1=phy_addr
# $2=g_led0_state
# $3=g_led1_state
# $4=g_led2_state
#-----------------------------------------------------
aqr_phy_led_set() {
    local G_LED0_ADDR="0x401EC430"
    local G_LED1_ADDR="0x401EC431"
    local G_LED2_ADDR="0x401EC432"
    local phy_addr=$1;
    local g_led0_state=$2;
    local g_led1_state=$3;
    local g_led2_state=$4;

    if [ $g_led0_state == "0" ]; then
        ssdk_sh debug phy set $phy_addr 0x401EC430 0x0;
    else
        ssdk_sh debug phy set $phy_addr 0x401EC430 0x100;
    fi

    if [ $g_led1_state == "0" ]; then
        ssdk_sh debug phy set $phy_addr 0x401EC431 0x0;
    else
        ssdk_sh debug phy set $phy_addr 0x401EC431 0x100;
    fi

    if [ $g_led2_state == "0" ]; then
        ssdk_sh debug phy set $phy_addr 0x401EC432 0x0;
    else
        ssdk_sh debug phy set $phy_addr 0x401EC432 0x100;
    fi
}


########################################################
# MAIN
########################################################
DBG "CPU: $CPU";
DBG "ethernet: $ETH";
DBG "#=$#; 1=$1; 2=$2, 3=$3, 4=$4"

case $CPU in
	"IPQ8070" | "IPQ8072" | "IPQ8074")
		case $ETH in
			"AQR111" | "AQR111C" | "AQR112" | "AQR112C" | "AQR112C_QCA8081")
                if [ $# == "4" ]; then
                    aqr_phy_led_set $1 $2 $3 $4;
                else
                    usage;
                    exit 0;
                fi
				;;
			*)
				echo "unknow ethernet type!";
				exit 1;
				;;
		esac
		;;
	*)
		echo "unknown CPU type!";
		exit 1;
		;;
esac
