#!/bin/sh
#
# Copyright (c) 2014, 2016, The Linux Foundation. All rights reserved.
#

. /lib/sn_functions.sh
. /lib/senao-shell-libs/network.sh

funcDebug=0
default_cfg80211=1

#added from SPF10.0_PRE_CS
qdss_tracing=0
daemon=0
cold_boot_support=0

wifi_profile="/lib/wifi/wifi_profile"
sn_wifi_profile="/lib/wifi/sn_wifi_profile"
#qca_wifi_profile="/lib/wifi/qca_wifi_profile"  #QCA configs, no need now

hostap_profile="/lib/wifi/hostapd_profile"
sn_hostap_profile="/lib/wifi/sn_hostapd_profile"

wifi_prof_config_name="wifiprofile"
hostap_prof_config_name="hostapprofile"

snConfig="snWifiConf"
snHostapConfig="snHostapConf"

wifi_prof_config="/etc/config/$wifi_prof_config_name"
hostap_prof_config="/etc/config/$hostap_prof_config_name"

DebugPrint(){
        [ "$funcDebug" == "1" ] && echo ">>>>>""$@" > /dev/console
}

wlanconfig() {
        [ "$funcDebug" == "1" ] && echo ">>>>>"wlanconfig "$@" > /dev/console
        /usr/sbin/wlanconfig "$@"
}

iwconfig() {
        [ "$funcDebug" == "1" ] && echo ">>>>>"iwconfig "$@" > /dev/console
        /usr/sbin/iwconfig "$@"
}

cfg80211tool() {
	[ "$funcDebug" == "1" ] && echo ">>>>>"cfg80211tool "$@" > /dev/console
        /usr/sbin/cfg80211tool "$@"
}

iwpriv() {
        [ "$funcDebug" == "1" ] && echo ">>>>>"iwpriv "$@" > /dev/console
        /usr/sbin/iwpriv "$@"
}

device_if() {
	if [ "$default_cfg80211" == "1" ]; then
		cfg80211tool "$@"
	else
		iwpriv "$@"
	fi
}

delay_cmd() {
	[ "$funcDebug" == "1" ] && echo sleep "$@" > /dev/console
        delay=$1
        shift
        ( sleep $delay && $@ ) &
}

insmod() {
	[ "$funcDebug" == "1" ] && echo insmod "$@" > /dev/console
	/usr/sbin/insmod "$@"
}

get_cpu_num() {
	local cpunum=`cat /proc/cpuinfo |grep ^processor |wc -l`
    echo ${cpunum:-1}
}

get_sn_wifi_option() {

	if [ ! -f "$wifi_prof_config" ]; then
	{
		if [ "$1" == "RELOAD_WIFI_KMODS" ]; then
			echo 0		#sometimes the config not prepared, we need return the RELOAD_WIFI_KMODS default value 0
		else
			echo 1		#if there is no wifi profile config, we assume all the options are 1 to avoid function failed
		fi
		return
	}
	fi
	option=$1
	
	ret=$(uci -q get $wifi_prof_config_name"."$snConfig"."$option);
	
	if [ "$ret" == "1" ]; then 
		echo 1
	else
		DebugPrint "$option not equal to 1"
		echo 0
	fi
}

get_sn_hostap_option() {

        if [ ! -f "$hostap_profile" ]; then
        {
                echo 1          #if there is no hostapd profile config, we assume all the options are 1 to avoid function failed
                echo "there is not get hostap profile config from $hostap_profile." > /dev/console
                return
        }
        fi
        option=$1

        ret=$(cat $hostap_profile | grep $option)

        if [ -z "$ret" ]; then
                echo 0
                return
        fi

        ret=$(cat $hostap_profile | grep $option | grep "#")

        if [ -z "$ret" ]; then
                echo 1
        else
                DebugPrint "$option not enabled"
                echo 0
        fi
}

addOption() {
        optionName=$1
        optionValue=$2
	prof_config=$3
        printf "\toption %-40s '%s'\n" $optionName $optionValue >> $prof_config
}

wifiProfile2Config(){
	if [ ! -s "$wifi_prof_config" ]; then
        #if wifi profile config already exists, no need do it
        #this config used for reference, the value should not be altered

	        if [ -s "$wifi_profile" ]; then

	                cat $wifi_profile | sed 's/^#.*//g' | grep "SENAO" > $sn_wifi_profile     #prepare sn_wifi_profile, it only contain SENAO_XXXX 

	                #cp $wifi_profile $qca_wifi_profile     
	                #sed -i '/SENAO/d' $qca_wifi_profile    #QCA configs, no need now
	        fi

        	if [ -s "$sn_wifi_profile" ]; then

			echo "config profile $snConfig" > $wifi_prof_config

                	lineNum=$(sed -n '$=' "$sn_wifi_profile")

                	for i in `seq 1 $lineNum`; do
                	        optionName=$(cat $sn_wifi_profile | head -n $i |tail -n 1 | awk -F 'SENAO_' '{print $2}'|awk -F '=' '{print $1}')
                	        optionValue=$(cat $sn_wifi_profile | head -n $i |tail -n 1 | awk -F 'SENAO_' '{print $2}'|awk -F '=' '{print $2}')
                	        [ "$optionValue" == "1" ] && addOption $optionName $optionValue $wifi_prof_config
                	done
			
			addOption "RELOAD_WIFI_KMODS" "0" $wifi_prof_config
		fi
	fi
}

HostapdProfile2Config(){
	if [ ! -s "$hostap_prof_config" ]; then

		if [ -s "$hostapd_profile" ]; then
			cat $hostap_profile | sed 's/^#.*//g' | grep "SENAO" > $sn_hostap_profile
		fi
		
		if [ -s "$sn_hostap_profile" ]; then

			echo "config profile $snHostapConfig" > $hostap_prof_config

			lineNum=$(sed -n '$=' "$sn_hostap_profile")
			
			for i in `seq 1 $lineNum`; do
				optionName=$(cat $sn_hostap_profile | head -n $i |tail -n 1 | awk -F 'SENAO_' '{print $2}'|awk -F '=' '{print $1}')
				optionValue=$(cat $sn_hostap_profile | head -n $i |tail -n 1 | awk -F 'SENAO_' '{print $2}'|awk -F '=' '{print $2}')
				[ "$optionValue" == "y" ] && addOption $optionName "1" $hostap_prof_config
			done
		fi

	fi 		
}

get_module_parameter() {
	moduleName=$1
	paramName=$2

	if [ -e /sys/module/$moduleName/parameters/$paramName ]; then
		value=$(cat /sys/module/$moduleName/parameters/$paramName)
	
		if [ -n $value ]; then
			echo $value
		else	
			echo 0
		fi
	else
		echo 0
	fi
}

hostapd_global_restart() {

        if  [ -e "/var/run/hostapd-global.pid" ]; then
                kill $(cat /var/run/hostapd-global.pid) &> /dev/null
                rm /var/run/hostapd-global.pid &> /dev/null
        
	        hostapd -g /var/run/hostapd/global -B -P /var/run/hostapd-global.pid
	fi
}

start_recovery_daemon() {
	[ "$funcDebug" == "1" ] && echo starting recovery daemon > /dev/console
	killall acfg_tool
	/usr/sbin/acfg_tool -e -s > /dev/console &
}

#update_qdss_tracing_daemon_variables() {
#	[ -f /sys/firmware/devicetree/base/soc_version_major ] && {
#		soc_version_major="$(hexdump -n 1 -e '"%1d"' /sys/firmware/devicetree/base/soc_version_major)"
#		if [ $soc_version_major = 1 ] && [ -f /ini/internal/QCA8074_i.ini ];then
#			qdss_tracing="$(grep enable_qdss_tracing /ini/internal/QCA8074_i.ini  | awk -F '=' '{print $2}')"
#			daemon="$(grep enable_daemon_support /ini/internal/QCA8074_i.ini  | awk -F '=' '{print $2}')"
#			qdss_tracing=$(($qdss_tracing && $daemon))
#		elif [ $soc_version_major = 2 ] && [ -f /ini/internal/QCA8074V2_i.ini ];then
#			qdss_tracing="$(grep enable_qdss_tracing /ini/internal/QCA8074V2_i.ini  | awk -F '=' '{print $2}')"
#			daemon="$(grep enable_daemon_support /ini/internal/QCA8074V2_i.ini  | awk -F '=' '{print $2}')"
#			qdss_tracing=$(($qdss_tracing && $daemon))
#		fi
#	}
#}

update_daemon_coldboot_qdss_support_variables() {
	local board_name=""
	local board_prefix=""
	local version=0
	local fw_ini_file=/lib/firmware/firmware_rdp_feature.ini

	[ ! -f $fw_ini_file  ] && {
		echo "******FW ini file not found******" >> /dev/console
		return
	}

	# Low Mem 256 profile does not support these
	[ -e /sys/firmware/devicetree/base/MP_256 ] && {
		echo "****** No Coldboot/QDSS Support for low mem ******" > /dev/console
		return
	}

	[ -f /tmp/sysinfo/board_name ] && {
		board_name=$(cat /tmp/sysinfo/board_name)
	}

	[ -z "$board_name"  ] && {
		echo "******Unable to find board name******" > /dev/console
		return
	}

	board_prefix=$(echo $board_name | sed 's/[0-9].*//g')

	case "$board_name" in
	ap-hk*)
		[ -f /sys/firmware/devicetree/base/soc_version_major ] && {
			version="$(hexdump -n 1 -e '"%1d"' /sys/firmware/devicetree/base/soc_version_major)"
		}

		local default_string=$board_prefix"_v"$version"_default"
		;;
	*)
		local default_string=$board_prefix"_default"
		;;
	esac

	#Grep the board_name in the fw_ini_file, if its not found, use the
	#default string
	boardname_grep_result=`grep -ci $board_name $fw_ini_file`
	default_grep_result=`grep -ci $default_string $fw_ini_file`

	if [ $boardname_grep_result != 0 ]
	then
		daemon=`grep -i $board_name $fw_ini_file | grep "enable_daemon_support" | awk -F'=' '{print $2}' | cut -c 1`
		cold_boot_support=`grep -i $board_name $fw_ini_file | grep "enable_cold_boot_support" | awk -F'=' '{print $2}' | cut -c 1`
		qdss_tracing=`grep -i $board_name $fw_ini_file | grep "enable_qdss_tracing" | awk -F'=' '{print $2}' | cut -c 1`
	elif [ $default_grep_result != 0 ]
	then
		daemon=`grep -i $default_string $fw_ini_file | grep "enable_daemon_support" | awk -F'=' '{print $2}' | cut -c 1`
		cold_boot_support=`grep -i $default_string $fw_ini_file | grep "enable_cold_boot_support" | awk -F'=' '{print $2}' | cut -c 1`
		qdss_tracing=`grep -i $default_string $fw_ini_file | grep "enable_qdss_tracing" | awk -F'=' '{print $2}' | cut -c 1`
	else
		echo "***** No coldboot or daemon support for $board_name******" > /dev/console
	fi

}


enable_qdss_tracing() {
	local board_name

	[ -f /tmp/sysinfo/board_name ] && {
		board_name=$(cat /tmp/sysinfo/board_name)
	}

	case "$board_name" in
	ap-hk*)
		[ ! -f /tmp/qdss_trace_commands ] && {
			echo "q6mem" > /sys/bus/coresight/devices/coresight-tmc-etr/out_mode
			echo 1 > /sys/bus/coresight/devices/coresight-tmc-etr/curr_sink
			echo "0x06021FB0 0xc5acce55" > /sys/bus/coresight/devices/coresight-hwevent/setreg
			echo "0x06130FB0 0xc5acce55" > /sys/bus/coresight/devices/coresight-hwevent/setreg
			echo "0x06021000 0x00000320" > /sys/bus/coresight/devices/coresight-hwevent/setreg
			echo "0x06130000 0x00000340" > /sys/bus/coresight/devices/coresight-hwevent/setreg
			echo 1 > /sys/bus/coresight/devices/coresight-stm/enable
			echo "*****Registers configuration for qdss_tracing completed*******" > /dev/console
		}
		echo -e "qdss_trace_load_config\nqdss_trace_start\nquit\n" > /tmp/qdss_trace_commands
		echo "******Starting cnss_cli********" > /dev/console
		cnsscli < /tmp/qdss_trace_commands > /dev/null
		;;
	*)
		echo "INFO:QDSS_tracing not applicable for $board_name" > /dev/console
		;;
	esac
}

qdss_post_setting() {
	[ $qdss_tracing = 1 ] && {
		#update_qdss_tracing_daemon_variables
		enable_qdss_tracing
	}
}

## pre-CAC
get_dfs_nol_list() {

    # only 5GHz supported
    [ "$1" == "wifi0" ] && return 0

    # only ap mode supported
    [ "$2" != "ap" ] && return 0

    if [ "$(uci -q get wireless.$1.pre_cac_en)" = "1" ]; then
        sn_zero_wait_dfs="1"
    else
    	#echo "******No sn_zero_wait_support*****"  > /dev/console
	return 0
    fi

    #echo -e "\033[44;37m [get_dfs_nol_list][$1][$2][$sn_zero_wait_dfs]\033[0m" > /dev/console
    if [ -f "/tmp/zerowaitdfs.lock" ]; then
        rm -f "/tmp/zerowaitdfs.lock"
    fi

    if [ -f "/tmp/zerowaitdfs.done" ]; then
	radartool -i "$1" getnol "/tmp/working.nol" > /dev/null 2>&1
    fi
}

set_dfs_nol_list() {

    # only 5GHz supported
    [ "$1" == "wifi0" ] && return 0

    # only ap mode supported
    [ "$2" != "ap" ] && return 0

    if [ "$(uci -q get wireless.$1.pre_cac_en)" = "1" ]; then
        sn_zero_wait_dfs="1"
    else
       	#echo "******No sn_zero_wait_support*****"  > /dev/console
	return 0 
    fi

    #echo -e "\033[44;37m [set_dfs_nol_list][$1][$2][$sn_zero_wait_dfs]\033[0m" > /dev/console
    if [ -f "/tmp/zerowaitdfs.done" ] && [ -f "/tmp/working.nol" ]; then
	# set nol file to working radio
	radartool -i "$1" usenol 0
	radartool -i "$1" setnol "/tmp/working.nol" > /dev/null 2>&1
	radartool -i "$1" usenol 1

	# set ignorecac to skip CAC time
	radartool -i "$1" ignorecac 1
    fi
}
## pre-CAC

do_cold_boot_calibration() {
	#Update the daemon, coldboot and qdss support variables from FW INI
	update_daemon_coldboot_qdss_support_variables

	[ $cold_boot_support = 0 ] && {
		echo "******No cold_boot_support*****"  > /dev/console
		return 0
	}

	[ -f /tmp/cold_boot_done ] && {
		echo "******Not the first boot. Skip coldboot calibration*****"  > /dev/console
		return 0
	}

	#daemon and cold_boot_support are assigned a value globally
	echo $cold_boot_support > /sys/module/cnss2/parameters/cold_boot_support
	echo $daemon > /sys/module/cnss2/parameters/daemon_support

	[ $daemon = 1 ] && {
		echo "*****starting cnssdaemon*****" > /dev/console
		/usr/bin/cnssdaemon &
	}
	echo "*********initiating cold boot calibration*************" > /dev/console
	coldbootmode=7
	#Set Cold boot mode to 10 for FTM Mode, 7 otherwise
	is_ftm="$(grep wifi_ftm_mode /proc/cmdline | wc -l)"
	[ $is_ftm = 1  ] && {
		coldbootmode=10
		echo "*****FTM Coldboot mode set to 10*****" > /dev/console
	}
	insmod qca_ol testmode=$coldbootmode

	pid=`pgrep cnssdaemon` && echo -1000 > /proc/$pid/oom_score_adj
	echo "*****cnssdaemon pid=$pid*********" > /dev/console
	# insmod and rmmod wifi_3_0 kernel object in addition to qca_ol
	# as wifi_3_0 initiates and shuts down driver.
	insmod wifi_3_0
	rmmod wifi_3_0
	rmmod qca_ol
	sync
	touch /tmp/cold_boot_done
}

dfs_spoof() {
	support_dfs_spoof=$(get_sn_wifi_option SUPPORT_DFS_SPOOF)
	[ "$support_dfs_spoof" == "0" ] && return

	local device="$1"

	[ "$device" == "wifi0" ] && return

	config_get dfs_ignore "$device" dfs_ignore 0
	device_if "$device" dfs_ignore "$dfs_ignore"

}

green_mode() {

        support_green_mode=$(get_sn_wifi_option SUPPORT_GREEN_MODE)
        [ "$support_green_mode" == "0" ] && return
        
        local device="$1"
        local domain=$(setconfig -g 4)
        local obey=$(uci get wireless.$device.obeyregpower)
        local country=$(uci get wireless.$device.country)
        local outdoor=$(uci -q get sysProductInfo.model.outdoor || echo -n "0")
        local dfsCertified=0

        # domain 0 = FCC
        # domain 1 = ETSI/EU
        # domain 2 = INT/others
        if [ "$obey" = "undefined" ]; then
                obey=1
                case ${domain} in
                        0)
                                country=840
                        ;;
                        1)
                                country=528
                        ;;
                        2)
                        ;;
                        *)
                        ;;
                esac
                uci set wireless.${device}.obeyregpower="$obey"
                uci set wireless.${device}.country="$country"
                uci commit wireless
        fi


	[ "$device" == "wifi0" ] && return

	if [ -f "/lib/wifi/RegularDomain.sh" ]; then
	{

	        local check_wifi2=$(/usr/sbin/foreach wireless wifi-device disabled 0| grep wifi2) #jacky added, to check the platform has wifi2 or not
	        local wifi2_opmode=$(uci -q get wireless.wifi2.opmode)

        	if [ "$wifi2_opmode" != "mon" -a -n "$check_wifi2" ]; then
                	local val=$(sh /lib/wifi/RegularDomain.sh $country $domain $obey $outdoor $dfsCertified $check_wifi2)
        	else
                	local val=$(sh /lib/wifi/RegularDomain.sh $country $domain $obey $outdoor $dfsCertified)
        	fi

	        local disable_band=$(echo $val | cut -d ' ' -f 1)
	        local weather_ch=$(echo $val | cut -d ' ' -f 2) # whether support weather radar channel
	
		support_dis_weather_ch=$(get_sn_wifi_option SUPPORT_DISABLE_WEATHER_RADAR_CHANNEL)
	
	        [ "$support_dis_weather_ch" == "0" ] && device_if "$device" weather_ch $weather_ch
	
	        [ -n "$check_wifi2" ] && {
	                local split_band=$(uci -q get wireless.$device.split_band) #Jacky:used for tri-band platform, this variable defined by vendor using
	
        	        case ${device} in
        	                "wifi1")
        	                        [ -z "$split_band" ] && split_band=12   #12 means band3+band4=4+8=12=high band, channel100~165, default tri-band setting
        	                        disable_band=$(($disable_band & $split_band))
        	                        ;;
        	                "wifi2")
        	                        [ -z "$split_band" ] && split_band=3    #3 means band1+band2=1+2=3=low band, channel36~64, default tri-band setting
        	                        disable_band=$(($disable_band & $split_band))
        	                        ;;
        	        esac
        	}

        	[ "$disable_band" == "0" ] && {
			config_set "$device" nochannel 1
        	        uci set wireless.$device.nochannel="1"
        	        uci commit wireless
        	} || {
			config_set "$device" nochannel 0
        	        uci set wireless.$device.nochannel="0"
        	        uci commit wireless
        	        device_if "$device" disable_band "$disable_band"
        	}
	}
	fi
}

TEMP_NOCTL_INFO_FILE="/tmp/noCtlFiles"

searchNextBoardData(){

        numOfSearch=$(cat $TEMP_NOCTL_INFO_FILE | grep $1 -c)

        if [ $numOfSearch -eq 1 ]; then
                noCTLBoardData=$(cat $TEMP_NOCTL_INFO_FILE | grep $1)
                sed -i '/'"$noCTLBoardData"'/d' $TEMP_NOCTL_INFO_FILE
                shift
                keyWordPriority=$@
                echo $noCTLBoardData
        elif [ $numOfSearch -eq 0 ]; then
                shift
                [ -n "$1" ] && searchNextBoardData "$@"
        else    #found more than 1 file with this keyword
                noCTLBoardData=$(cat $TEMP_NOCTL_INFO_FILE | grep $1 | grep $2 -c)
                shift
                searchNextBoardData "$@"
        fi
}

check_boarddata() {
    local wifi_dev=$1
    local FIRMWARE_PATH="/lib/firmware/"
    local keyWordPriority="2G wifi0 IPQ4019 5G wifi1" #key word search priority
                                                                           
    ls -R $FIRMWARE_PATH | grep noCTL > $TEMP_NOCTL_INFO_FILE
                                                                           
    numOfNoCTL=$(ls -R $FIRMWARE_PATH | grep -c noCTL)
                                                                                                   
    for count in `seq 0 $(($numOfNoCTL-1))`; do                                            
                               
	noCTLBData=$(searchNextBoardData $keyWordPriority)

	[ -z "$noCTLBData" ] && continue

	CTLBData=$(echo $noCTLBData | awk -F '_noCTL' '{print $1}')".bin"
	noCTLBDataPath=$(find $FIRMWARE_PATH -name $noCTLBData)
	CTLBDataPath=$(echo $noCTLBDataPath | awk -F '_noCTL' '{print $1}')".bin"

	drv_load_data=$CTLBDataPath

	obey=$(uci get wireless.$wifi_dev.obeyregpower)
        [ "$obey" == "undefined" ] && obey=1

	if [ "$obey" == "1" ]; then
		rom_boarddata="/rom${CTLBDataPath}"
	else
		rom_boarddata="/rom${noCTLBDataPath}"
	fi

        [ -n "$rom_boarddata" ] && [ -n "$drv_load_data" ] && ln -sf $rom_boarddata $drv_load_data
        DebugPrint "rom_boarddata=$rom_boarddata, drv_load_data=$drv_load_data"
    done  

    [ -f "$TEMP_NOCTL_INFO_FILE" ] && rm $TEMP_NOCTL_INFO_FILE -f
}

prepare_ax_boarddata(){

	local BOARDDATA_PATH="/lib/firmware/IPQ8074/"
	local DRIVER_BDATA_NAME="bdwlan.bin"
	local SENAO_BDATA_CTL_NAME="bdwlan_sn_CTL.bin"	#this file should be copied from sn_files/bdwlan.bxx when compiling package senao-ipq807x-wifi-fw-image
	local SENAO_BDATA_NOCTL_NAME="bdwlan_sn_noCTL.bin"
	local SENAO_BDATA_IC_NAME="bdwlan_sn_IC.bin"

	# cypress
	grep -q IPQ6018 /tmp/sysinfo/model && BOARDDATA_PATH="/lib/firmware/IPQ6018/"

	[ -f /sys/firmware/devicetree/base/soc_version_major ] && {
	    soc_version_major="$(hexdump -n 1 -e '"%1d"' /sys/firmware/devicetree/base/soc_version_major)"
	    if [ "$soc_version_major" == "2" ]; then
		if [ -f /proc/device-tree/soc/wifi1@c0000000/qcom,board_id ]; then
			board_id=$(hexdump -n 2 -s 2 -C /proc/device-tree/soc/wifi1@c0000000/qcom,board_id | awk '{ print $2$3}' | cut -b 2,3,4)
			DRIVER_BDATA_NAME="bdwlan.b"$board_id
		fi
	    else
		if [ -f /proc/device-tree/soc/wifi@c0000000/qcom,board_id ]; then
			board_id=$(hexdump -n 2 -s 2 -C /proc/device-tree/soc/wifi@c0000000/qcom,board_id | awk '{ print $2$3}' | cut -b 3,4)
			DRIVER_BDATA_NAME="bdwlan.b"$board_id
		fi
		# cypress
		if [ -f /proc/device-tree/soc/wifi@c000000/qcom,board_id ]; then
			board_id=$(hexdump -n 2 -s 2 -C /proc/device-tree/soc/wifi@c000000/qcom,board_id | awk '{ print $2$3}' | cut -b 3,4)
			DRIVER_BDATA_NAME="bdwlan.b"$board_id
		fi
	    fi
	    DebugPrint "soc ver=$soc_version_major, DRIVER_BDATA_NAME=$DRIVER_BDATA_NAME"
	}

	local DRIVER_LOAD_DATA="${BOARDDATA_PATH}${DRIVER_BDATA_NAME}"
	local target_load_data=""

	local previous_load_data=$(ls -al $DRIVER_LOAD_DATA | awk -F '-> ' '{print $2}')

	if [ ! -e $DRIVER_LOAD_DATA ]; then
		DebugPrint "===$DRIVER_LOAD_DATA not exist==="
		return
	fi

	#obey set 1 if anyone radio is 1
	local apObey=$(/usr/sbin/foreach wireless wifi-device obeyregpower 1)
	if [ -z "$apObey" ]; then
		apObey=$(/usr/sbin/foreach wireless wifi-device obeyregpower undefined)
		if [ -z "$apObey" ]; then
			apObey=0
		else
			apObey=1
		fi
	else
		apObey=1
	fi

	if [ "$apObey" == "0" ]; then
		target_load_data="${BOARDDATA_PATH}${SENAO_BDATA_NOCTL_NAME}"
	else
		config_get country "wifi0" country #just get wifi0 country since all radio should be in the same country
		if [ "$country" == "124" ]; then #CANADA country code 124
			target_load_data="${BOARDDATA_PATH}${SENAO_BDATA_IC_NAME}"
		else
			target_load_data="${BOARDDATA_PATH}${SENAO_BDATA_CTL_NAME}"
		fi
	fi

	if [ -n "$target_load_data" -a -e "$target_load_data" ]; then

		if [ "$previous_load_data" != "$target_load_data" ]; then
			DebugPrint "=== change $previous_load_data to $target_load_data, unload_qcawifi ==="
			unload_qcawifi
			ln -sf $target_load_data $DRIVER_LOAD_DATA
			DebugPrint "=== $DRIVER_LOAD_DATA -> $target_load_data ==="
		fi
	else
		DebugPrint "target boarddata $target_load_data not exist"
	fi
}

sn_bandsteering_check() {
    # We only need to check a couple SSID at least,   #
    # even if we support multiple SSID band-steering. #
    # SSID-1 band-steering check #

    support_band_steering=$(get_sn_wifi_option SUPPORT_BAND_STEERING_WITH_SUPPRESS_PROBE_RESP_AND_ASSOCIATION)

    [ "$support_band_steering" == "0" ] && return

    local id1=1 # wifi0 vap index
    local id2=1 # wifi1 vap index

    # check vap enable #
    wifi0_ssid_disabled=$(uci get wireless.wifi0_ssid_"$id1".disabled)
    wifi1_ssid_disabled=$(uci get wireless.wifi1_ssid_"$id2".disabled)
    if [ "$wifi0_ssid_disabled" = "1" ] || [ "$wifi1_ssid_disabled" = "1" ]; then
        DebugPrint "radio disable, bandsteer 0"
        bandsteer=0
        return
    fi

    # check SSID #
    wifi0_ssid=$(uci get wireless.wifi0_ssid_"$id1".ssid)
    wifi1_ssid=$(uci get wireless.wifi1_ssid_"$id2".ssid)
    if [ "$wifi0_ssid" != "$wifi1_ssid" ]; then
        DebugPrint "SSID mismatch, bandsteer 0"
        bandsteer=0
        return
    fi

    # check encryption #
    wifi0_ssid_enc=$(uci get wireless.wifi0_ssid_"$id1".encryption)
    wifi1_ssid_enc=$(uci get wireless.wifi1_ssid_"$id2".encryption)
    if [ "$wifi0_ssid_enc" != "$wifi1_ssid_enc" ]; then
        DebugPrint "encryption mismatch, bandsteer 0"
        bandsteer=0
        return
    fi

    # check wep key #
    if [ "${wifi0_ssid_enc:0:3}" = "wep" ]; then
        wifi0_ssid_key_id=$(uci get wireless.wifi0_ssid_"$id1".key_id)
        wifi1_ssid_key_id=$(uci get wireless.wifi1_ssid_"$id1".key_id)
        if [ "$wifi0_ssid_key_id" != "$wifi1_ssid_key_id" ]; then
            DebugPrint "wep key id mismatch, bandsteer 0"
            bandsteer=0
            return
        fi

        wifi0_ssid_key=$(uci get wireless.wifi0_ssid_"$id1".key"$wifi0_ssid_key_id")
        wifi1_ssid_key=$(uci get wireless.wifi1_ssid_"$id2".key"$wifi1_ssid_key_id")
        if [ "$wifi0_ssid_key" != "$wifi1_ssid_key" ]; then
            DebugPrint "wep key mismatch, bandsteer 0"
            bandsteer=0
            return
        fi
    fi

    # check wpa-psk key #
    if [ "${wifi0_ssid_enc:0:3}" = "psk" ]; then
        wifi0_ssid_key=$(uci get wireless.wifi0_ssid_"$id1".key)
        wifi1_ssid_key=$(uci get wireless.wifi1_ssid_"$id2".key)
        if [ "$wifi0_ssid_key" != "$wifi1_ssid_key" ]; then
            DebugPrint "psk key mismatch, bandsteer 0"
            bandsteer=0
            return
        fi
    fi
}

cal_ack_time()
{
	local distance="$1"

	# The Tx Ack value ranges from 64 to 255
	#
	# slottime = 9 + (distance / 300) + ((distance % 300) ? 1 : 0);
	# acktimeout = slottime * 2 + 3;
	#

	if [ $(( $distance % 300 )) -gt 0 ]; then
		local slottime=$(( ( ( $distance / 300 ) + 9) + 1 ))
	else
		local slottime=$(( ( ( $distance / 300 ) + 9) ))
	fi

	local ack=$(( ( $slottime *2 ) + 3 ))
	[ $ack -gt 255 ] && ack="255"
	[ $ack -lt 64 ] && ack="64"

	echo "$ack"
}


##### wifi distance #####
distance_func()
{
	local device="$1"
	local phydevice="$2"

	DebugPrint "distance_func device:$device phydevice:$phydevice"
	config_get hwmode "$device" hwmode	
	config_get distance "$device" distance

	[ -n "$distance" ] && {
		#	local w5g_dfs_recovery1="echo $distance > /proc/sys/dev/$phydevice/distance"
		#	echo "$distance" > /proc/sys/dev/$phydevice/distance

		local ack=$(cal_ack_time $distance)
		device_if "$device" acktimeout "$ack"
	}
	enable_dfs_check "$device" "$hwmode"
}

enable_dfs_check() 
{
	#config_get hwmode "$device" hwmode
	local device="$1"
	local hwmode="$2"
	case "$hwmode" in
		*a|11na|11ac)
			#enable_dfs_check $device  " $w5g_dfs_recovery1"
			[ -f "$testcase" ] &&
			dfs_func "$device" "$w5g_dfs_recovery1" &
		;;
		*) DebugPrint "No define for DFS check:hwmode_dfs=$hwmode" ;;
	esac
}

qcawifi_start_hostapd_cli() {
	local device=$1
	local ifidx=0
	local radioidx=${device#wifi}

	config_get vifs $device vifs

	[ -z $vifs ] && vifs=$(eval "/usr/sbin/foreach wireless wifi-iface device $device")

	for vif in $vifs; do
		local config_methods vifname

		config_get vifname "$vif" ifname

		if [ -n $vifname ]; then
			[ $ifidx -gt 0 ] && vifname="ath${radioidx}$ifidx" || vifname="ath${radioidx}"
		fi

		config_get_bool wps_pbc "$vif" wps_pbc 0
		config_get config_methods "$vif" wps_config
		[ "$wps_pbc" -gt 0 ] && append config_methods push_button

		if [ -n "$config_methods" ]; then
			pid=/var/run/hostapd_cli-$vifname.pid
			hostapd_cli -i $vifname -P $pid -a /lib/wifi/wps-hostapd-update-uci -p /var/run/hostapd-$device -B
		fi

		ifidx=$(($ifidx + 1))
	done
}

enable_fastscan() {

	support_fastscan=$(get_sn_wifi_option SUPPORT_BACKGROUND_FASTSCAN)

	[ "$support_fastscan" == "0" ] && return	

	echo 1 > /tmp/notify_fastscan_enable

	triband5G_fastscan_status=0;  #0:both disabled, 1: only wifi1 enable, 2: only wifi2 , 3 both enabled
	fastscan_script_24G="/sbin/Fastscan24G.sh"
	fastscan_script_5G="/sbin/Fastscan5G.sh"
	isConfigured=$(uci -q get apcontroller.capwap.enable)
	test "$(pgrep -f Fastscan)" && kill -9 $(pgrep -f Fastscan)
	if [ "$isConfigured" = "1" ];then
		DebugPrint "AP controlled by controller, do fastscan by capwap"
	else
		fastscan_duration_24G=20
		fastscan_duration_5G=20

		local vifs=$(eval "/usr/sbin/foreach wireless wifi-iface device wifi0")
		for vif in $vifs; do
			if [ "$(uci -q get wireless.$vif.fastroamingEnable)" == "1" ]; then
				[ -e "$fastscan_script_24G" ] && sh "$fastscan_script_24G" $fastscan_duration_24G &
				break
			fi
		done

		local vifs=$(eval "/usr/sbin/foreach wireless wifi-iface device wifi1")
		for vif in $vifs; do
			if [ "$(uci -q get wireless.$vif.fastroamingEnable)" == "1" ]; then
				triband5G_fastscan_status=$(expr $triband5G_fastscan_status + 1)
				break
			fi
		done

		local vifs=$(eval "/usr/sbin/foreach wireless wifi-iface device wifi2")
		for vif in $vifs; do
			if [ "$(uci -q get wireless.$vif.fastroamingEnable)" == "1" ]; then
				triband5G_fastscan_status=$(expr $triband5G_fastscan_status + 2)
				break
			fi
		done
		if [ "$triband5G_fastscan_status" == "0" ]; then
			DebugPrint "No need to open 5G fastscan"
		else
			[ -e "$fastscan_script_5G" ] && sh "$fastscan_script_5G"  $fastscan_duration_5G $triband5G_fastscan_status &
		fi
	fi
	
}

disable_fastscan() {

        support_fastscan=$(get_sn_wifi_option SUPPORT_BACKGROUND_FASTSCAN)

        [ "$support_fastscan" == "0" ] && return

	echo 0 > /tmp/notify_fastscan_enable
}

set_dis_legacy_rate() {
	ifname=$1
	minRate=$2
	case "$minRate" in
		2)
		rateCode=0x0001
		bcn_rate=2000
		#device_if $ifname set_bcn_rate 2000
		;;
		5)	#should be 5.5, but for convenience, avoid use the point
		rateCode=0x0003
		bcn_rate=5500
		#device_if $ifname set_bcn_rate 5500
		;;
		6)
		rateCode=0x0007
		bcn_rate=6000
		#device_if $ifname set_bcn_rate 6000
		;;
		9)
		rateCode=0x0017
		bcn_rate=9000
		#device_if $ifname set_bcn_rate 9000
		;;
		11)
		rateCode=0x0037
		bcn_rate=11000
		#device_if $ifname set_bcn_rate 11000
		;;
		12)
		rateCode=0x003f
		bcn_rate=12000
		#device_if $ifname set_bcn_rate 12000
		;;
		18)
		rateCode=0x007f
		bcn_rate=18000
		#device_if $ifname set_bcn_rate 18000
		;;
		24)
		rateCode=0x00ff
		bcn_rate=24000
		#device_if $ifname set_bcn_rate 24000
		;;
	esac
	device_if $ifname dis_legacy $rateCode
	###Due to 5G bit rate set fail####
	device_if $ifname set_bcn_rate $bcn_rate
	
}

set_airtime_fairness_vip() {
	local check_atf=$(uci -q get airtime_fairness.atf)     #jacky added, to check the venfor has airtime_fairness or not
	if [ "$check_atf" == "atf" ]; then
		atf_enabled=$(uci -q get airtime_fairness.atf.enabled)
		atf_usage=$(uci -q get airtime_fairness.atf.usage)
		[ -e "/etc/crontabs/root" -a -n "$(cat /etc/crontabs/root | grep updateATFConfig)" ] && {
			sed -i '/updateATFConfig/d' /etc/crontabs/root
		}
		if [ "$atf_usage" == "vip" ]; then      ##maybe add more usage in the future
			if [ "$atf_enabled" == "1" ]; then
				[ -e "/tmp/set_vip_list" ] && rm /tmp/set_vip_list -f #remove this because wifi reload and atf setting also reset
				uci set airtime_fairness.atf.parameter_changed=1
				uci commit airtime_fairness
				echo "*/1 * * * * /bin/sh /sbin/updateATFConfig.sh &" >> /etc/crontabs/root
				crontab /etc/crontabs/root
			fi
		fi
	fi
}

set_vlan_passthrough() {

        support_vlanpass=$(get_sn_wifi_option SUPPORT_VLAN_PASSTHROUGH)

        [ "$support_vlanpass" == "0" ] && return

	local mesh_off=$(uci -q get mesh.wifi.disabled)
	if [ ! -e "/lib/wifi/vlan_passthrough.sh" ]; then
		DebugPrint "/lib/wifi/vlan_passthrough.sh not exists"
		return 0
	fi

	vlan_passthrough
}

create_dummy_interfaces() {

	local wifi_dev=$1
	local dummyList="/tmp/dummyIf_"$wifi_dev	#since some vendor use ath29 as repeater interface, use this file to indicate whether it is dummy or not

	[ -e "$dummyList" ] && rm "$dummyList" -f

	local athw=$(ifconfig 2>/dev/null|grep $wifi_dev);

	if [ "$default_cfg80211" == "1" ]; then
		local athw=$(cat /sys/class/net/$wifi_dev/carrier)
	else
		local athw=$(ifconfig 2>/dev/null|grep $wifi_dev);
	fi

	athdummy=$(uci get wireless."$wifi_dev"_dummy.ifname)

	[ -z "$athdummy" ] && { #default dummy interface of wifix
	    athdummy="dummy"$(echo $wifi_dev| awk -F 'wifi' '{print$2}')
	}

        [ -z "$athw" ] && {
		if [ "$default_cfg80211" == "1" ]; then
			ifconfig $wifi_dev up
			wlanconfig "$athdummy" create wlandev "$wifi_dev" wlanmode "ap" nosbeacon -cfg80211
			iw phy "$(cat /sys/class/net/$wifi_dev/phy80211/name)" interface add $athdummy type __ap
		else
	                wlanconfig "$athdummy" create wlandev "$wifi_dev" wlanmode "ap" nosbeacon
		fi
		echo $athdummy >> $dummyList
        }
}

destroy_dummy_interfaces() {

	local wifi_dev=$1
	athdummy=$(uci get wireless."$wifi_dev"_dummy.ifname)
	[ -z "$athdummy" ] && { #default dummy interface of wifix
            athdummy="dummy"$(echo $wifi_dev| awk -F 'wifi' '{print$2}')
        }
	destroy_vap $athdummy
}
set_min_rate() {
	#  SENAO support disable legacy rate (set min rate)
	local wifi_dev=$1
	local ifname=$2
	config_get minLegacyRate "$wifi_dev" min_rate 0
        [ $minLegacyRate -gt 1 ] && [ $minLegacyRate -le 24 ] && {
                if [ $minLegacyRate -le 6 ] && [[ "$wifi_dev" == "wifi1" || "$wifi_dev" == "wifi2" ]]; then
                        DebugPrint "5G device $device min data rate is 6Mbps, cannot disable lower rates"
                else
                        set_dis_legacy_rate $ifname $minLegacyRate
                        if [ "$wifi_dev" == "wifi0" ] && [ $minLegacyRate -gt 11 ]; then
                        	DebugPrint "disabled all b mode basic rate, set pureg=1 for basic rate setting"
                                device_if $ifname pureg 1
                        fi
                fi
        }
}

set_client_limit() {
	#sENAO support client limit function

        support_connlimit=$(get_sn_wifi_option SUPPORT_CONNECTION_LIMIT)

        #[ "$support_connlimit" == "0" ] && return

	local wifi_dev=$1
	config_get clientlimits_enable "$wifi_dev" clientlimits_enable
        config_get clientlimits_number "$wifi_dev" clientlimits_number
        
        [ -n $clientlimits_enable ] && [ $clientlimits_enable -eq 0 ] && {
		clientlimits_number=$(uci -q get functionlist.vendorlist.CONNECTION_LIMIT_NUM)
		[ -z $clientlimits_number ] && clientlimits_number=127
	}

        [ -n $clientlimits_number ] && {
		device_if $wifi_dev max_radio_sta $clientlimits_number
		if [ "$support_connlimit" == "1" ]; then
		    case "$wifi_dev" in
                        wifi0)
                        echo "$clientlimits_number" > /proc/connection_limit
                        ;;
                        wifi1)
                        echo "$clientlimits_number" > /proc/connection_limit_5g
                        ;;
                        wifi2)
                        echo "$clientlimits_number" > /proc/connection_limit_5g_2
                        ;;
                    esac
		fi
        }
}

get_beacon_interval() {
	# beacon interval. get bintval(default value 100ms) here, not from wireless cfg file
	#just preset the beacon interval, the device_if command will be done later
	local wifi_dev=$1
	config_get vifs "$wifi_dev" vifs
	config_get opmode "$wifi_dev" opmode
	local bintval=100  # set default 100ms
	local working_vap_num=0

	[ -z $vifs ] && vifs=$(eval "/usr/sbin/foreach wireless wifi-iface device $wifi_dev")

	for vif in $vifs; do
		config_get opmode "$device" opmode
		config_get mode "$vif" mode
		config_get disabled "$vif" disabled
		config_get mode_display "$vif" mode_display
		config_get mode_display_2 "$vif" mode_display_2
		config_get mesh "$vif" mesh 0
		local mesh_disabled=$(uci -q get mesh.wifi.disabled)
		config_get mgmt_vif "$vif" mgmt_vif 0

		[ "$disabled" == "1" ] && continue;

		# check mode_display/mode_display_2 without mesh / mgmt intf
		[ $mesh -eq 0 -a $mgmt_vif -eq 0 ] && [ "$opmode" != "$mode_display" ] && [ "$opmode" != "$mode_display_2" ] && continue;

		# Not create mesh intf when mesh_disabled = 1
		[ $mesh -eq 1 ] && [ ${mesh_disabled:-0} -eq 1 ] && continue;

		# Not create mgmt interface in controller mode.
		[ $mgmt_vif -eq 1 ] && [ $(uci get apcontroller.capwap.enable) -eq 1 ] && continue;

		#monitor mode interface will not send beacon, not to count in working vap
		[ "$mode" == "monitor" ] && continue

		working_vap_num=$(expr $working_vap_num + 1)
	done
	# if vap number of ap,mesh,and guest network in total are more than 8, set beacon interval  to 200ms
	if [ $working_vap_num -gt 8 ]; then
		bintval=200
	fi
	echo $bintval
}

set_beacon_interval() {

        wifi_dev=$1
        beacon_int=$(get_beacon_interval "$wifi_dev")

        if [ -n "$lastApIf" ]; then
                device_if $lastApIf bintval $beacon_int
        else
                radioIdx=$(echo $wifi_dev | awk -F "wifi" '{print $2}')
                ifname=$(/sbin/getWifiFirstIF "$radioIdx")
                device_if $ifname bintval $beacon_int
        fi
}

set_sn_bandsteering() {

    local wifi_dev=$1
    # force use 1st SSID
    local ifname="$(uci get wireless.$wifidev_ssid_1.ifname)"
    config_get bandsteer "$wifi_dev" bandsteer

    [ -n "$bandsteer" ] && echo "$bandsteer" > /proc/bandsteer

    if [ -e "/sbin/sn_bandsteering" ]; then
	/sbin/sn_bandsteering           #new flow, create a bandsteer checking script
    elif [ "$bandsteer" != "0" ]; then
	local firstApIfname="$(uci get wireless.$wifidev_ssid_1.ifname)"
	local check_wifi2=$(/usr/sbin/foreach wireless wifi-device disabled 0| grep wifi2) #jacky added, to check the platform has wifi2 or not
	config_get bandsteer "$wifi_dev" bandsteer
	config_get bandsteerrssi "$wifi_dev" bandsteerrssi
	config_get bandsteerpersent "$wifi_dev" bandsteerpersent
	config_get bandsteerHBrssi "$wifi_dev" bandsteerHBrssi
	config_get bandsteerHBpersent "$wifi_dev" bandsteerHBpersent

	sn_bandsteering_check      

	config_set bandsteer "$wifi_dev" bandsteer

	[ -n "$bandsteer" ] && echo "$bandsteer" > /proc/bandsteer

	[ -n "$bandsteer" -a "$bandsteer" != "0" ] && {
	    [ "$ifname" == $firstApIfname ] && {
		device_if "$ifname" bs_en 1
	    }
	}
	[ -n "$bandsteer" -a "$bandsteer" = "2" ] && {
	    [ "$ifname" == $firstApIfname ] && {
		device_if "$ifname" bsmin5grssi "$bandsteerrssi"
		[ -n "$check_wifi2" ] && {
		    device_if "$ifname" bs5ghbrssi  "$bandsteerHBrssi"
		}
	    }
	}
	[ -n "$bandsteer" -a "$bandsteer" = "3" ] && {
	    [ "$ifname" == $firstApIfname ] && {
		device_if "$ifname" bsmin5grssi "$bandsteerrssi"
		device_if "$ifname" bsloadratio "$bandsteerpersent"
		[ -n "$check_wifi2" ] && {
		    device_if "$ifname" bs5ghbrssi  "$bandsteerHBrssi"
		    device_if "$ifname" bshbratio "$bandsteerHBpersent"
		}
	    }
	}
    fi
}

set_sn_bandsteering_per_ssid() {

    local device=$1
    local ap_vifs=$2
    file="/tmp/indexfile"

    #config_get bandsteer "$wifi_dev" bandsteer
    bandsteer=1

    [ -n "$bandsteer" ] && echo "$bandsteer" > /proc/bandsteer

    # get the enabled ssid index
    touch $file
    for vif in $ap_vifs; do
	case "$vif" in
            *ssid*)
		index="$(echo $vif |awk -F '_' '{print $3}')"
		getindex=$(cat $file |grep $index)
		if [ -z $(cat $file |grep $index) ]; then
                    echo $index >>$file
		fi
		;;
	esac
    done

    if [ "$device" = "wifi0" ]; then
	if [ -e "/sbin/sn_bandsteering" ]; then
	    exec < $file
	    while read line
	    do
		/sbin/sn_bandsteering $line
	    done
	    rm $file
	fi
    fi
}

set_sn_fasthandover() {

        support_fasthandover=$(get_sn_wifi_option SUPPORT_FAST_HANDOVER_INDEPENDENT)

        [ "$support_fasthandover" == "0" ] && return
	
	local wifi_dev=$1
	config_get opmode "$wifi_dev" opmode
        [ "$opmode" != "ap" ] &&  fasthandover_status="0" || config_get fasthandover_status "$wifi_dev" fasthandover_status
        config_get fasthandover_rssi "$wifi_dev"   fasthandover_rssi

        if [ "$fasthandover_rssi" -lt -95 ]; then
                fasthandover_rssi=-95
        fi
        case "$wifi_dev" in
                wifi0)
                echo "$fasthandover_status" "$fasthandover_rssi" > /proc/ap_roaming
                ;;
                wifi1)
                echo "$fasthandover_status" "$fasthandover_rssi" > /proc/ap_roaming_5g
                ;;
                wifi2)
                echo "$fasthandover_status" "$fasthandover_rssi" > /proc/ap_roaming_5g_2
                ;;
        esac
}

set_macfilter() {
	
	vif=$1
	ifname=$2
	config_get macfilter "$vif" macfilter
	case "$macfilter" in
	allow)
		config_get allowmaclist "$vif" allowmaclist
		[ -n "$allowmaclist" ] && {
			device_if "$ifname" maccmd 3 # flush MAC list
			while [ -n "$allowmaclist" ]; do
				mactmp=`echo $allowmaclist | cut -d ' ' -f 1`
				device_if "$ifname" addmac "$mactmp"
				allowmaclist=`echo $allowmaclist|cut -c 18-`
			done
		}
		device_if "$ifname" maccmd 1
	;;
	deny)
		config_get denymaclist "$vif" denymaclist
		[ -n "$denymaclist" ] && {
			device_if "$ifname" maccmd 3 # flush MAC list
			while [ -n "$denymaclist" ]; do
				mactmp=`echo $denymaclist | cut -d ' ' -f 1`
				device_if "$ifname" addmac "$mactmp"
				denymaclist=`echo $denymaclist|cut -c 18-`
			done
		}
		device_if "$ifname" maccmd 2
	;;
	*)
	# default deny policy if mac list exists
	#[ -n "$denymaclist" ] && device_if "$ifname" maccmd 2
	# default disable and flush MAC list
	device_if "$ifname" maccmd 3
	device_if "$ifname" maccmd 0
	;;
	esac
}

create_wifi_vap(){

	device=$1
	vif=$2
	config_get opmode $device opmode	
	local nosbeacon= wlanaddr=""
        config_get mode "$vif" mode
        local wlanmode=$mode
        [ "$default_cfg80211" == "1" ] && pmode=$mode
        config_get enc "$vif" encryption "none"
        config_get mode_display "$vif" mode_display
        config_get mode_display_2 "$vif" mode_display_2
        config_get mesh "$vif" mesh 0
        local mesh_disabled=$(uci -q get mesh.wifi.disabled)
        config_get mgmt_vif "$vif" mgmt_vif 0

        # check mode_display/mode_display_2 without mesh / mgmt intf
        [ $mesh -eq 0 -a $mgmt_vif -eq 0 ] && [ "$opmode" != "$mode_display" ] && [ "$opmode" != "$mode_display_2" ] && continue;

        # Not create mesh intf when mesh_disabled = 1
        [ $mesh -eq 1 ] && [ ${mesh_disabled:-0} -eq 1 ] && continue;

        # Not create mgmt interface in controller mode.
        [ $mgmt_vif -eq 1 ] && [ $(uci get apcontroller.capwap.enable) -eq 1 ] && continue;
	config_get ifname "$vif" ifname
        config_get_bool nosbeacon "$device" nosbeacon
        config_get qwrap_enable "$device" qwrap_enable 0
        all_ifnames=$(echo $all_ifnames $ifname)

        if [ "$default_cfg80211" == "1" ]; then
	        [ "$wlanmode" = "ap" ] && wlanmode="__ap"
                [ "$wlanmode" = "sta" ] && wlanmode="managed"
                [ "$wlanmode" = "lite_monitor" ] && wlanmode="monitor"
                [ "$wlanmode" = "ap_monitor" ] && wlanmode="__ap"
                [ "$wlanmode" = "ap_smart_monitor" ] && wlanmode="__ap"
                [ "$wlanmode" = "ap_lp_iot" ] && wlanmode="__ap"
                [ "$wlanmode" = "mesh" ] && wlanmode="__ap"
                [ "$wlanmode" = "wrap" ] && wlanmode="__ap"

                [ "$pmode" = "ap_monitor" ] && pmode="specialvap"
                [ "$pmode" = "ap_smart_monitor" ] && pmode="smart_monitor"
                [ "$pmode" = "ap_lp_iot" ] && pmode="lp_iot_mode"
        fi
	[ "$nosbeacon" = 1 ] || nosbeacon=""

        if [ "$default_cfg80211" == "1" ]; then
        	local checkDevUp=$(ifconfig 2>/dev/null|grep $device);
                [ -z "$checkDevUp" ] && ifconfig $device up
	fi

        if [ "$mode" = "sta" -a "$qwrap_enable" -gt 0 ]; then
        	wlanaddr="00:00:00:00:00:00"
                if [ "$default_cfg80211" == "1" ]; then
                	wlanconfig "$ifname" create wlandev "$device" wlanmode "$pmode" ${wlanaddr:+wlanaddr "$wlanaddr"} ${nosbeacon:+nosbeacon} -cfg80211
                        DebugPrint "cfg80211: ifname: $ifname mode: $wlanmode cfgphy: $(cat /sys/class/net/$device/phy80211/name)"
                        iw phy "$(cat /sys/class/net/$device/phy80211/name)" interface add $ifname type $wlanmode
                        iw "$ifname" set 4addr on >/dev/null 2>&1
                else
                        DebugPrint "wlanconfig $ifname create wlandev $device wlanmode $wlanmode ${wlanaddr:+wlanaddr $wlanaddr} ${nosbeacon:+nosbeacon}"
                        ifname=$(/usr/sbin/wlanconfig "$ifname" create wlandev "$device" wlanmode "$wlanmode" ${wlanaddr:+wlanaddr "$wlanaddr"} ${nosbeacon:+nosbeacon})
                fi
	else
                if [ "$default_cfg80211" == "1" ]; then
                	wlanconfig "$ifname" create wlandev "$device" wlanmode "$pmode" ${nosbeacon:+nosbeacon} -cfg80211
                        DebugPrint "cfg80211: ifname: $ifname mode: $wlanmode cfgphy: $(cat /sys/class/net/$device/phy80211/name)"
                        iw phy "$(cat /sys/class/net/$device/phy80211/name)" interface add $ifname type $wlanmode
                else
                	DebugPrint "wlanconfig $ifname create wlandev $device wlanmode $mode ${nosbeacon:+nosbeacon}"
                        ifname=$(/usr/sbin/wlanconfig "$ifname" create wlandev "$device" wlanmode "$mode" ${nosbeacon:+nosbeacon})
		fi
	fi

        [ $? -ne 0 ] && {
        	DebugPrint "enable_qcawifi($device): Failed to set up $mode vif $ifname"
                return
        }
        config_set "$vif" ifname "$ifname"
}

create_wifi_vaps(){
        device=$1

        config_get vifs "$device" vifs

        for vif in $vifs; do
		create_wifi_vap $device $vif
	done
}

destroy_vap() {
	local ifname="$1"
	ifconfig $ifname down
	wlanconfig $ifname destroy
}

set_vap_params() {

	device=$1
	vif=$2
	
	config_get opmode "$device" opmode
	local start_hostapd= start_wpa_supplicant= nosbeacon= wlanaddr=""
        config_get mode "$vif" mode
        local wlanmode=$mode
        local is_valid_owe_group=0
        local is_valid_sae_group=0
        config_get enc "$vif" encryption "none"
        config_get mode_display "$vif" mode_display
        config_get mode_display_2 "$vif" mode_display_2
        config_get mesh "$vif" mesh 0
        local mesh_disabled=$(uci -q get mesh.wifi.disabled)
        config_get mgmt_vif "$vif" mgmt_vif 0
        config_get_bool hs20 "$vif" hs20 0

        # check mode_display/mode_display_2 without mesh / mgmt intf
        [ $mesh -eq 0 -a $mgmt_vif -eq 0 ] && [ "$opmode" != "$mode_display" ] && [ "$opmode" != "$mode_display_2" ] && return;
        # Not create mesh intf when mesh_disabled = 1
        [ $mesh -eq 1 ] && [ ${mesh_disabled:-0} -eq 1 ] && return;

        # Not create mgmt interface in controller mode.
        [ $mgmt_vif -eq 1 ] && [ $(uci get apcontroller.capwap.enable) -eq 1 ] && return;

        if [ "$mode" == "ap" ]; then
		vap_ap_mode_setting "$device" "$vif"
		ap_vifs="$vif $ap_vifs"
	else
		vap_non_ap_mode_setting "$device" "$vif"
	fi

	vap_common_setting "$device" "$vif"
	[ "$hs20" -gt 0 ] && vap_hotspot_setting "$device" "$vif"
	#vap_non_using_setting "$device" "$vif"         #non using parameters
	
	#2020.07.10 Senao George, fix cypress long distance throughput issue.
	local outdoor=$(uci get sysProductInfo.model.outdoor)
	grep -q IPQ6018 /tmp/sysinfo/model && [ $outdoor -eq 1 ] && vap_distance_setting "$device" "$vif"
	
	#2020.3.20 Senao George, for wifi debug
	config_get ifname "$vif" ifname
	config_get dbgLVL "$vif" dbgLVL
	[ -n "$dbgLVL" ] && device_if "$ifname" dbgLVL "$dbgLVL"

    #2020.10.30 Senao George, porting from qca_spf_11_0_csu3
    config_get ext_nss_sup "$vif" ext_nss_sup
	case "$ext_nss_sup" in
	        1|on|enabled) device_if "$ifname" ext_nss_sup 1 >/dev/null 2>&1
	    	;;
	    	0|on|enabled) device_if "$ifname" ext_nss_sup 0 >/dev/null 2>&1
	    	;;
	    	*) ;;
	esac
}

set_wifi_param_per_vap() {

	device=$1
        local ap_vifs

	config_get vifs "$device" vifs
	config_get opmode "$device" opmode
	for vif in $vifs; do
		set_vap_params $device $vif
        done   	#end of device vifs loop

	if [ "$opmode" == "ap" ]; then
		#SENAO bandsteering setting
		support_band_steering=$(get_sn_wifi_option SUPPORT_BAND_STEERING_WITH_SUPPRESS_PROBE_RESP_AND_ASSOCIATION)
		if [ "$support_band_steering" != "0" ]; then
			band_steering_per_ssid=$(get_sn_wifi_option SUPPORT_BAND_STEERING_PER_SSID)
			if [ "$band_steering_per_ssid" == "1" ]; then
				set_sn_bandsteering_per_ssid "$device" "$ap_vifs"
			else
				set_sn_bandsteering $device
			fi
		fi
	fi

}

remove_hostapd_conf() {
	dev=$1
	[ -f /sys/class/net/${dev}/parent ] && { \
		local parent=$(cat /sys/class/net/${dev}/parent)
		[ -n "$parent" -a "$parent" = "$device" ] && { \
			#new flow, to kill wifi related processes which use wpa_cli and global
			[ -f "/var/run/hostapd-${dev}.lock" ] && { \
				wpa_cli -g /var/run/hostapd/global raw REMOVE ${dev}
				rm /var/run/hostapd-${dev}.lock
			}
		}
	}
}

set_aggregation() {

	wifi_dev=$1
	ifname=$2

	config_get aggregation_enable "$wifi_dev" aggregation_enable

	if [ -n "$aggregation_enable" ]; then
		if [ "$aggregation_enable" == "1" ]; then
			config_get aggregation_frame "$wifi_dev" aggregation_frame
			config_get amsdu_frame "$wifi_dev" amsdu_frame
			[ -n "$aggregation_frame" ] && device_if "$ifname" ampdu "$aggregation_frame"
			[ -n "$amsdu_frame" ] && device_if "$ifname" amsdu "$amsdu_frame"
			local power=0;
			config_get aggregation_byte "$wifi_dev" aggregation_byte
			[ -n "$aggregation_byte" ] && {
				if [ "$aggregation_byte" -gt  "8191" ] && [ "$aggregation_byte" -le  "16383" ]; then
					power=1;
				elif [ "$aggregation_byte" -gt  "16383" ] && [ "$aggregation_byte" -le  "32767" ]; then
					power=2;
				elif [ "$aggregation_byte" -gt  "32767" ] && [ "$aggregation_byte" -le  "65535" ]; then
					power=3;
				else
					power=0;
				fi
				device_if "$ifname" maxampdu "$power"
			}
		else
			device_if "$ifname" ampdu "64"
			device_if "$ifname" maxampdu "3"
		fi
	fi
}

fix_mpsk_vlan_net() {
	local device=$1
	local vif=$2
	local ifname=$3
	local vlanid=""
	local MvlanID=$(uci -q get network.sys.ManagementVLANID)
	local MvlanEnable=$(uci -q get network.sys.WLANVLANEnable)

	config_get wpa_psk_file "$vif" wpa_psk_file

	if [ -f "$wpa_psk_file" ]
	then
		. /lib/network/config.sh
		for vlanid in `cat $wpa_psk_file| awk -F':' '{print $1}'| awk -F'vlanid=' '{print $2}'| awk '{print $1}'`
		do
			if [ -d /sys/class/net/br-vlan$vlanid ]
			then
				setup_interface "$ifname.$vlanid" "vlan$vlanid"
			elif [ "${MvlanEnable:-0}" = "1" -a "$MvlanID" = "$vlanid" ]
			then
				setup_interface "$ifname.$vlanid" "lan"
			fi
		done
	fi
}

vap_common_setting() {	#parameters for ap and sta mode

	device=$1
	vif=$2
	local mpsk_no_vlan=$(uci -q get functionlist.functionlist.MPSK_NOT_SUPPORT_VLAN)

	config_get opmode "$device" opmode
	config_get ifname "$vif" ifname

	set_hwmode_htmode "$ifname" "$device" "$opmode"

        config_get wds "$vif" wds
        case "$wds" in
                 1|on|enabled) wds=1;;
                 *) wds=0;;
        esac
        device_if "$ifname" wds "$wds"

	config_get shortgi "$vif" shortgi 1
	config_get hwmode "$device" hwmode
	if [ "$hwmode" == "11axa" -o "$hwmode" == "11axg" ]; then	#jacky: since spf10.0_pre_cs will not block 11axa with 400ns, we block the non-standard setting in script
		shortgi=0
	fi
        [ -n "$shortgi" ] && device_if "$ifname" shortgi "${shortgi}"    

        config_get multi_group_key "$vif" multi_group_key
        [ -n "$multi_group_key" ] && device_if "$ifname" multi_group_key $multi_group_key

        config_get_bool uapsd "$vif" uapsd 1
        [ -n "$uapsd" ] && device_if "$ifname" uapsd "$uapsd"

        config_get  backhaul "$vif" backhaul 0
        device_if "$ifname" backhaul "$backhaul" >/dev/null 2>&1

        config_get mcast_rate "$vif" mcast_rate
        [ "$mcast_rate" ] && device_if "$ifname" mcast_rate "${mcast_rate}"

        config_get mcastenhance "$vif" mcastenhance
        [ -n "$mcastenhance" ] && {
                [ "$mcastenhance" -gt "0" ] && mcastenhance=6
                device_if "$ifname" mcastenhance "${mcastenhance}"
        }

        config_get bcmcdrop "$vif" bcmcdrop
        [ -n "$bcmcdrop" ] && device_if "$ifname" bcmcdrop "${bcmcdrop}"

        config_get metimer "$vif" metimer
        [ -n "$metimer" ] && device_if "$ifname" metimer "${metimer}"

        config_get metimeout "$vif" metimeout
        [ -n "$metimeout" ] && device_if "$ifname" metimeout "${metimeout}"

        config_get_bool medropmcast "$vif" medropmcast
        [ -n "$medropmcast" ] && device_if "$ifname" medropmcast "${medropmcast}"

        config_get me_adddeny "$vif" me_adddeny
        [ -n "$me_adddeny" ] && device_if "$ifname" me_adddeny ${me_adddeny}

        config_get_bool stafwd "$vif" stafwd 0
        [ -n "$stafwd" ] && device_if "$ifname" stafwd "$stafwd"

	config_get he_ulofdma "$vif" he_ulofdma 0 # SENAO disable uplink ofdma, if u want to open it, set it in wireleess config
        [ -n "$he_ulofdma" ] && device_if "$ifname" he_ulofdma "${he_ulofdma}"

	if [ "$mode_display" == "mesh" ];then
	    device_if "$ifname" he_dlofdma 0 # SENAO disable downlink ofdma at mesh interface.
	fi

        ### SENAO ### 11k fastroaming-- preset 11k disable, sw spec requirement
        #device_if "$ifname" rrm 0
        if [ "$mode_display" == "ap" -o "$mode_display" == "wds_ap" ]; then
	    device_if "$ifname" apchanrpt 1  ## SENAO: always enable ap chan report ##
	    device_if "$ifname" rrm 1  ## SENAO: always enable 802.11k ##
        fi

        device_if "$ifname" wmm 1  #no matter AP/STA mode, it should be enabled

        device_if "$ifname" doth 1 #no matter AP/STA mode, it should be enabled

        config_get nss "$vif" nss
        [ -n "$nss" -a "$ac_mode" == "1" ] && device_if "$ifname" nss "$nss"

        #ap_isoloation, it seems only wrap mode no need to use this
        config_get_bool ap_isolation_enabled $device ap_isolation_enabled 0
        config_get_bool isolate "$vif" isolate 0

        if [ $ap_isolation_enabled -ne 0 ]; then
		[ "$mode" = "wrap" ] && isolate=1
        fi
        local net_cfg bridge
        net_cfg="$(find_net_config "$vif")"
        [ -z "$net_cfg" -o "$isolate" = 1 -a "$mode" = "wrap" ] || {
		[ -f /sys/class/net/${ifname}/parent ] && { \
			bridge="$(bridge_interface "$net_cfg")"
			config_set "$vif" bridge "$bridge"
			start_net "$ifname" "$net_cfg"
			[ "$multi_group_key" = "1" -a "$mpsk_no_vlan" != "1" ] && fix_mpsk_vlan_net "$device" "$vif" "$ifname"
		}
        }
	#end of ap_isolation

        config_get ssid "$vif" ssid
        [ -n "$ssid" ] && {
		if [ "$default_cfg80211" == "1" ]; then
			cfg80211tool "$ifname" ssid  "$ssid"
		else
		        iwconfig "$ifname" essid "$ssid"
		fi
		#iwconfig "$ifname" essid on
		#iwconfig "$ifname" essid ${ssid:+-- }"$ssid"
        }

        set_wifi_up "$vif" "$ifname"

	common_vap_setting_after_vap_up "$vif"

        set_txpower "$device" "$vif"

        set_hostapd_wpa_supplicant "$vif"

        set_aggregation "$device" "$ifname"

        #  support traffic control
        [ -e "/lib/wifi/traffic_control.sh" ] && tc_setConf "$vif"

	# support disable legacy rate (set min rate)
	config_get nawds_vif $vif nawds
	if [ "$nawds_vif" != "1" ]; then
	    set_min_rate $device $ifname
	fi
}

vap_ap_mode_setting() {	#parameters for ap mode only

	device=$1
	vif=$2
	config_get ifname "$vif" ifname

	config_get_bool hidden "$vif" hidden 0
	[ -n "$hidden" ] && device_if "$ifname" hide_ssid "$hidden"

	config_get_bool dynamicbeacon "$vif" dynamicbeacon 0
	[ $hidden = 1 ] && device_if "$ifname" dynamicbeacon "$dynamicbeacon"

	#jacky: not to set beacon interval here, move to set_beacon_interval and just set in lastApIf or getWifiFirstIF
	# beacon interval. get bintval(default value 100ms) here, not from wireless cfg file.
	#local bintval=$(get_beacon_interval "$device")
	#[ -n "$bintval" ] && {
        #device_if "$ifname" bintval "$bintval"
        #config_set "$vif" bintval "$bintval"
        #}

	config_get dtim_period "$vif" dtim_period
	[ -n "$dtim_period" ] && device_if "$ifname" dtim_period "$dtim_period"

	config_get preamble "$vif" preamble
	[ -n "$preamble" ] && device_if "$ifname" shpreamble "$preamble"

	config_get_bool uapsd "$vif" uapsd 1
	[ -n "$uapsd" ] && device_if "$ifname" uapsd "$uapsd"

	config_get athnewind "$vif" athnewind
	[ -n "$athnewind" ] && device_if "$ifname" athnewind "$athnewind"
	
        config_get  backhaul "$vif" backhaul 0
        device_if "$ifname" backhaul "$backhaul" >/dev/null 2>&1


	config_get mcast_rate "$vif" mcast_rate
	[ "$mcast_rate" ] && device_if "$ifname" mcast_rate "${mcast_rate}"

	config_get mcastenhance "$vif" mcastenhance
	[ -n "$mcastenhance" ] && {
		[ "$mcastenhance" -gt "0" ] && mcastenhance=6
		device_if "$ifname" mcastenhance "${mcastenhance}"
	}

	config_get bcmcdrop "$vif" bcmcdrop
	[ -n "$bcmcdrop" ] && device_if "$ifname" bcmcdrop "${bcmcdrop}"

	config_get me_adddeny "$vif" me_adddeny
	[ -n "$me_adddeny" ] && device_if "$ifname" me_adddeny ${me_adddeny}
	#device_if "$ifname" me_adddeny 3758096635 255 255 255 #filter Bonjour/mDNS packet 
	#emr3000 use this, but the proper usage should set these parameter in wireless config, ex: wireless.wifi0_ssid_1.me_adddeny='3758096635 255 255 255'

	#per vap setting functions

	# black/white list setting
	set_macfilter $vif $ifname

	# support disable legacy rate (set min rate)
	#set_min_rate $device $ifname

	#use this for sn_set_channel_final, the channel setting command must used on ap interface
	lastApIf=$ifname

	config_get_bool proxyarp "$vif" proxyarp
	[ -n "$proxyarp" ] && device_if "$ifname" proxyarp "$proxyarp"
}

vap_non_ap_mode_setting() { #parameters for sta mode only

	device=$1
	vif=$2

	config_get ifname "$vif" ifname
	local extap=0
	local supp_mgmt=$(uci -q get functionlist.functionlist.SUPPORT_MANAGEMENT_SSID)

	# wds_sta extap set 1 when support mgmt ssid, it fix mgmt ssid can't up.
	#[ "$mode" = "sta" ] && [ "$mode_display" != "wds_sta" -o ${supp_mgmt:-0} -eq 1 ] && [ "$mode_display" != "sta_ap" ] && extap=1

	[ "$mode" = "sta" ] && [ "$mode_display" != "wds_sta" ] && [ "$mode_display" != "sta_ap" ] && extap=1

	if [ ${supp_mgmt:-0} -eq 1 ]; then 
	    local mgmt_radio=$(uci -c /rom/etc/config/ get functionlist.vendorlist.MANAGEMENT_SUPPORT_RADIO)
	    local wifi0_opmode=$(uci -q get wireless.wifi0.opmode)
	    local wifi1_opmode=$(uci -q get wireless.wifi1.opmode)

	    [ "$mgmt_radio" = '1' ] && [ "$wifi0_opmode" = "wds_sta" ] && extap=1
	    [ "$mgmt_radio" = '2' ] && [ "$wifi1_opmode" = "wds_sta" ] && extap=1
	fi
	echo -e "\033[41;37mextap[$extap]\033[0m" > /dev/console

	config_set "$vif" extap "$extap"
	[ -n "$extap" ] && device_if "$ifname" extap "$extap"
	[ -n "$extap" ] && [ "$default_cfg80211" == "1" ] && iw "$ifname" set 4addr on >/dev/null 2>&1
	## a monitor script to check wds-stp connection status ##
	if [ "$mode_display" = "wds_sta" ]; then
	    echo "*/1 * * * * /bin/sh /sbin/reconnect_wds_ap.sh &" >> /etc/crontabs/root
	    crontab /etc/crontabs/root
	fi

}

vap_distance_setting() {
	device=$1
	vif=$2

	config_get distance "$device" distance
	config_get ifname "$vif" ifname

	[ -n "$distance" ] && {
		local ack=$(cal_ack_time $distance)
		DebugPrint "QCA_BUG_FIX vap_distance_setting device:$device vif:$vif ifname:$ifname ack:$ack"
		wifitool $ifname setUnitTestCmd 0x48 2 317 $ack
	}
}

vap_hotspot_setting() {	#parameters for ap mode

	device=$1
	vif=$2

	config_get ifname "$vif" ifname

	config_get_bool l2tif "$vif" l2tif
	[ -n "$l2tif" ] && device_if "$ifname" l2tif "$l2tif"

	config_get_bool qbssload "$vif" qbssload
	[ -n "$qbssload" ] && device_if "$ifname" qbssload "$qbssload"

	config_get_bool proxyarp "$vif" proxyarp
	[ -n "$proxyarp" ] && device_if "$ifname" proxyarp "$proxyarp"

	config_get_bool dgaf_disable "$vif" dgaf_disable
	[ -n "$dgaf_disable" ] && device_if "$ifname" dgaf_disable "$dgaf_disable"

	config_get osen "$vif" osen
	[ -n "$osen" ] && device_if "$ifname" osen "$osen"

	config_get vhtmubfer "$vif" vhtmubfer
	[ -n "$vhtmubfer" ] && device_if "$ifname" vhtmubfer "$vhtmubfer"

	config_get hcbssload "$vif" hcbssload
	[ -n "$hcbssload" ] && device_if "$ifname" hcbssload "$hcbssload"

}

#parameters but not using now since it does not exist in our etc/config/wireless
#NOTE: if you find any parameter want to use in this function, please move to another function ex: vap_ap_mode_setting
vap_non_using_setting() { 

	device=$1	
	vif=$2
	
	config_get ifname "$vif" ifname

        config_get cfreq2 "$vif" cfreq2
        [ -n "$cfreq2" -a "$htmode" = "HT80_80" ] && device_if "$ifname" cfreq2 "$cfreq2"

	config_get_bool hidden "$vif" hidden 0

        config_get_bool dynamicbeacon "$vif" dynamicbeacon 0
        [ $hidden = 1 ] && device_if "$ifname" dynamicbeacon "$dynamicbeacon"

        config_get db_rssi_thr "$vif" db_rssi_thr
        [ -n "$db_rssi_thr" ] && device_if "$ifname" db_rssi_thr "$db_rssi_thr"

        config_get db_timeout "$vif" db_timeout
        [ -n "$db_timeout" ] && device_if "$ifname" db_timeout "$db_timeout"

        config_get nrshareflag "$vif" nrshareflag
        [ -n "$nrshareflag" ] && device_if "$ifname" nrshareflag "$nrshareflag"

        config_get ext_nss "$device" ext_nss
        case "$ext_nss" in
		1|on|enabled) device_if "$phy" ext_nss 1 >/dev/null 2>&1;;
                0|off|disabled) device_if "$phy" ext_nss 0 >/dev/null 2>&1;;
                *) ;;
        esac

        config_get ext_nss_sup "$vif" ext_nss_sup
        case "$ext_nss_sup" in
		1|on|enabled) device_if "$ifname" ext_nss_sup 1 >/dev/null 2>&1;;
		0|off|disabled) device_if "$ifname" ext_nss_sup 0 >/dev/null 2>&1;;
                *) ;;
        esac

        config_get TxBFCTL "$vif" TxBFCTL
        [ -n "$TxBFCTL" ] && device_if "$ifname" TxBFCTL "$TxBFCTL"

        config_get_bool countryie "$vif" countryie
        [ -n "$countryie" ] && device_if "$ifname" countryie "$countryie"

		config_get_bool vap_contryie "$vif" vap_contryie
		[ -n "$vap_contryie" ] && device_if "$ifname" vap_contryie "$vap_contryie"

	config_get ppdu_duration "$device" ppdu_duration
	[ -n "$ppdu_duration" ] && device_if "$phy" ppdu_duration "${ppdu_duration}"

	config_get he_ul_ppdu_dur "$device" he_ul_ppdu_dur
	[ -n "$he_ul_ppdu_dur" ] && device_if "$phy" he_ul_ppdu_dur "${he_ul_ppdu_dur}"

        config_get own_ie_override "$vif" own_ie_override
        [ -n "$own_ie_override" ] && device_if "$ifname" rsn_override 1

        config_get powersave "$vif" powersave
        [ -n "$powersave" ] && device_if "$ifname" powersave "${powersave}"

        config_get_bool ant_ps_on "$vif" ant_ps_on
        [ -n "$ant_ps_on" ] && device_if "$ifname" ant_ps_on "${ant_ps_on}"

        config_get ps_timeout "$vif" ps_timeout
        [ -n "$ps_timeout" ] && device_if "$ifname" ps_timeout "${ps_timeout}"

        config_get disable11nmcs "$vif" disable11nmcs
        [ -n "$disable11nmcs" ] && device_if "$ifname" disable11nmcs "${disable11nmcs}"

        config_get conf_11acmcs "$vif" conf_11acmcs
        [ -n "$conf_11acmcs" ] && device_if "$ifname" conf_11acmcs "${conf_11acmcs}"

        #support independent repeater mode
        config_get vap_ind "$vif" vap_ind
        [ -n "$vap_ind" ] && device_if "$ifname" vap_ind "${vap_ind}"

        config_get scanband "$vif" scanband
        [ -n "$scanband" ] && device_if "$ifname" scanband "${scanband}"

        config_get periodicScan "$vif" periodicScan
        [ -n "$periodicScan" ] && device_if "$ifname" periodicScan "${periodicScan}"

        config_get cwmin "$vif" cwmin
        [ -n "$cwmin" ] && device_if "$ifname" cwmin ${cwmin}

        config_get cwmax "$vif" cwmax
        [ -n "$cwmax" ] && device_if "$ifname" cwmax ${cwmax}

        config_get aifs "$vif" aifs
        [ -n "$aifs" ] && device_if "$ifname" aifs ${aifs}

        config_get txoplimit "$vif" txoplimit
        [ -n "$txoplimit" ] && device_if "$ifname" txoplimit ${txoplimit}

        config_get noackpolicy "$vif" noackpolicy
        [ -n "$noackpolicy" ] && device_if "$ifname" noackpolicy ${noackpolicy}

        config_get_bool wmm "$vif" wmm
        [ -n "$wmm" ] && device_if "$ifname" wmm "$wmm"

        config_get_bool doth "$vif" doth
        [ -n "$doth" ] && device_if "$ifname" doth "$doth"

        config_get doth_chanswitch "$vif" doth_chanswitch
        [ -n "$doth_chanswitch" ] && device_if "$ifname" doth_chanswitch ${doth_chanswitch}

        config_get quiet "$vif" quiet
        [ -n "$quiet" ] && device_if "$ifname" quiet "$quiet"

        config_get mfptest "$vif" mfptest
        [ -n "$mfptest" ] && device_if "$ifname" mfptest "$mfptest"

        config_get dtim_period "$vif" dtim_period
        [ -n "$dtim_period" ] && device_if "$ifname" dtim_period "$dtim_period"

        config_get noedgech "$vif" noedgech
        [ -n "$noedgech" ] && device_if "$ifname" noedgech "$noedgech"

        config_get ps_on_time "$vif" ps_on_time
        [ -n "$ps_on_time" ] && device_if "$ifname" ps_on_time "$ps_on_time"

        config_get inact "$vif" inact
        [ -n "$inact" ] && device_if "$ifname" inact "$inact"

        config_get wnm "$vif" wnm
        [ -n "$wnm" ] && device_if "$ifname" wnm "$wnm"

        config_get bpr_enable  "$vif" bpr_enable
        [ -n "$bpr_enable" ] && device_if "$ifname" set_bpr_enable "$bpr_enable"

        config_get vhtmaxampdu "$vif" vhtmaxampdu
        [ -n "$vhtmaxampdu" ] && device_if "$ifname" vhtmaxampdu "$vhtmaxampdu"

        config_get setaddbaoper "$vif" setaddbaoper
        [ -n "$setaddbaoper" ] && device_if "$ifname" setaddbaoper "$setaddbaoper"

        config_get addbaresp "$vif" addbaresp
        [ -n "$addbaresp" ] && device_if "$ifname" $addbaresp

        config_get addba "$vif" addba
        [ -n "$addba" ] && device_if "$ifname" addba $addba

        config_get delba "$vif" delba
        [ -n "$delba" ] && device_if "$ifname" delba $delba

        config_get vht_mcsmap "$vif" vht_mcsmap
        [ -n "$vht_mcsmap" ] && device_if "$ifname" vht_mcsmap "$vht_mcsmap"

        config_get he_mcs "$vif" he_mcs
        [ -n "$he_mcs" ] && device_if "$ifname" he_mcs "$he_mcs"

        config_get chbwmode "$vif" chbwmode
        [ -n "$chbwmode" ] && device_if "$ifname" chbwmode "$chbwmode"

	config_get ldpc "$vif" ldpc
	[ -n "$ldpc" ] && device_if "$ifname" ldpc "$ldpc"

	config_get rx_stbc "$vif" rx_stbc
	[ -n "$rx_stbc" ] && device_if "$ifname" rx_stbc "$rx_stbc"

	config_get tx_stbc "$vif" tx_stbc
	[ -n "$tx_stbc" ] && device_if "$ifname" tx_stbc "$tx_stbc"

	config_get cca_thresh "$vif" cca_thresh
	[ -n "$cca_thresh" ] && device_if "$ifname" cca_thresh "$cca_thresh"

	config_get set11NRetries "$vif" set11NRetries
	[ -n "$set11NRetries" ] && device_if "$ifname" set11NRetries "$set11NRetries"

	config_get chanbw "$vif" chanbw
	[ -n "$chanbw" ] && device_if "$ifname" chanbw "$chanbw"

	config_get maxsta "$vif" maxsta
	[ -n "$maxsta" ] && device_if "$ifname" maxsta "$maxsta"

	config_get sko_max_xretries "$vif" sko_max_xretries
	[ -n "$sko_max_xretries" ] && device_if "$ifname" sko "$sko_max_xretries"

	config_get extprotmode "$vif" extprotmode
	[ -n "$extprotmode" ] && device_if "$ifname" extprotmode "$extprotmode"

	config_get extprotspac "$vif" extprotspac
	[ -n "$extprotspac" ] && device_if "$ifname" extprotspac "$extprotspac"

	config_get_bool cwmenable "$vif" cwmenable
	[ -n "$cwmenable" ] && device_if "$ifname" cwmenable "$cwmenable"

	config_get_bool protmode "$vif" protmode
	[ -n "$protmode" ] && device_if "$ifname" protmode "$protmode"

	config_get enablertscts "$vif" enablertscts
	[ -n "$enablertscts" ] && device_if "$ifname" enablertscts "$enablertscts"

	config_get txcorrection "$vif" txcorrection
	[ -n "$txcorrection" ] && device_if "$ifname" txcorrection "$txcorrection"

	config_get rxcorrection "$vif" rxcorrection
	[ -n "$rxcorrection" ] && device_if "$ifname" rxcorrection "$rxcorrection"

        config_get vsp_enable "$vif" vsp_enable
        [ -n "$vsp_enable" ] && device_if "$ifname" vsp_enable "$vsp_enable"

        config_get txqueuelen "$vif" txqueuelen
        [ -n "$txqueuelen" ] && ifconfig "$ifname" txqueuelen "$txqueuelen"

        config_get tdls "$vif" tdls
        [ -n "$tdls" ] && device_if "$ifname" tdls "$tdls"

        config_get set_tdls_rmac "$vif" set_tdls_rmac
        [ -n "$set_tdls_rmac" ] && device_if "$ifname" set_tdls_rmac "$set_tdls_rmac"

        config_get tdls_qosnull "$vif" tdls_qosnull
        [ -n "$tdls_qosnull" ] && device_if "$ifname" tdls_qosnull "$tdls_qosnull"

        config_get tdls_uapsd "$vif" tdls_uapsd
        [ -n "$tdls_uapsd" ] && device_if "$ifname" tdls_uapsd "$tdls_uapsd"

        config_get tdls_set_rcpi "$vif" tdls_set_rcpi
        [ -n "$tdls_set_rcpi" ] && device_if "$ifname" set_rcpi "$tdls_set_rcpi"

        config_get tdls_set_rcpi_hi "$vif" tdls_set_rcpi_hi
        [ -n "$tdls_set_rcpi_hi" ] && device_if "$ifname" set_rcpihi "$tdls_set_rcpi_hi"

        config_get tdls_set_rcpi_lo "$vif" tdls_set_rcpi_lo
        [ -n "$tdls_set_rcpi_lo" ] && device_if "$ifname" set_rcpilo "$tdls_set_rcpi_lo"

        config_get tdls_set_rcpi_margin "$vif" tdls_set_rcpi_margin
        [ -n "$tdls_set_rcpi_margin" ] && device_if "$ifname" set_rcpimargin "$tdls_set_rcpi_margin"

        config_get tdls_dtoken "$vif" tdls_dtoken
        [ -n "$tdls_dtoken" ] && device_if "$ifname" tdls_dtoken "$tdls_dtoken"

        config_get do_tdls_dc_req "$vif" do_tdls_dc_req
        [ -n "$do_tdls_dc_req" ] && device_if "$ifname" do_tdls_dc_req "$do_tdls_dc_req"

        config_get tdls_auto "$vif" tdls_auto
        [ -n "$tdls_auto" ] && device_if "$ifname" tdls_auto "$tdls_auto"

        config_get tdls_off_timeout "$vif" tdls_off_timeout
        [ -n "$tdls_off_timeout" ] && device_if "$ifname" off_timeout "$tdls_off_timeout"

        config_get tdls_tdb_timeout "$vif" tdls_tdb_timeout
        [ -n "$tdls_tdb_timeout" ] && device_if "$ifname" tdb_timeout "$tdls_tdb_timeout"

        config_get tdls_weak_timeout "$vif" tdls_weak_timeout
        [ -n "$tdls_weak_timeout" ] && device_if "$ifname" weak_timeout "$tdls_weak_timeout"

        config_get tdls_margin "$vif" tdls_margin
        [ -n "$tdls_margin" ] && device_if "$ifname" tdls_margin "$tdls_margin"

	config_get tdls_rssi_ub "$vif" tdls_rssi_ub
        [ -n "$tdls_rssi_ub" ] && device_if "$ifname" tdls_rssi_ub "$tdls_rssi_ub"

        config_get tdls_rssi_lb "$vif" tdls_rssi_lb
        [ -n "$tdls_rssi_lb" ] && device_if "$ifname" tdls_rssi_lb "$tdls_rssi_lb"

        config_get tdls_path_sel "$vif" tdls_path_sel
        [ -n "$tdls_path_sel" ] && device_if "$ifname" tdls_pathSel "$tdls_path_sel"

        config_get tdls_rssi_offset "$vif" tdls_rssi_offset
        [ -n "$tdls_rssi_offset" ] && device_if "$ifname" tdls_rssi_o "$tdls_rssi_offset"

        config_get tdls_path_sel_period "$vif" tdls_path_sel_period
        [ -n "$tdls_path_sel_period" ] && device_if "$ifname" tdls_pathSel_p "$tdls_path_sel_period"

        config_get tdlsmacaddr1 "$vif" tdlsmacaddr1
        [ -n "$tdlsmacaddr1" ] && device_if "$ifname" tdlsmacaddr1 "$tdlsmacaddr1"

        config_get tdlsmacaddr2 "$vif" tdlsmacaddr2
        [ -n "$tdlsmacaddr2" ] && device_if "$ifname" tdlsmacaddr2 "$tdlsmacaddr2"

        config_get tdlsaction "$vif" tdlsaction
        [ -n "$tdlsaction" ] && device_if "$ifname" tdlsaction "$tdlsaction"

        config_get tdlsoffchan "$vif" tdlsoffchan
        [ -n "$tdlsoffchan" ] && device_if "$ifname" tdlsoffchan "$tdlsoffchan"

        config_get tdlsswitchtime "$vif" tdlsswitchtime
        [ -n "$tdlsswitchtime" ] && device_if "$ifname" tdlsswitchtime "$tdlsswitchtime"

        config_get tdlstimeout "$vif" tdlstimeout
        [ -n "$tdlstimeout" ] && device_if "$ifname" tdlstimeout "$tdlstimeout"

        config_get tdlsecchnoffst "$vif" tdlsecchnoffst
        [ -n "$tdlsecchnoffst" ] && device_if "$ifname" tdlsecchnoffst "$tdlsecchnoffst"

        config_get tdlsoffchnmode "$vif" tdlsoffchnmode
        [ -n "$tdlsoffchnmode" ] && device_if "$ifname" tdlsoffchnmode "$tdlsoffchnmode"

        config_get_bool blockdfschan "$vif" blockdfschan
        [ -n "$blockdfschan" ] && device_if "$ifname" blockdfschan "$blockdfschan"

        config_get dbgLVL "$vif" dbgLVL
        [ -n "$dbgLVL" ] && device_if "$ifname" dbgLVL "$dbgLVL"

        config_get dbgLVL_high "$vif" dbgLVL_high
        [ -n "$dbgLVL_high" ] && device_if "$ifname" dbgLVL_high "$dbgLVL_high"

        config_get acsmindwell "$vif" acsmindwell
        [ -n "$acsmindwell" ] && device_if "$ifname" acsmindwell "$acsmindwell"

        config_get acsmaxdwell "$vif" acsmaxdwell
        [ -n "$acsmaxdwell" ] && device_if "$ifname" acsmaxdwell "$acsmaxdwell"

        config_get acsreport "$vif" acsreport
        [ -n "$acsreport" ] && device_if "$ifname" acsreport "$acsreport"

        config_get ch_hop_en "$vif" ch_hop_en
        [ -n "$ch_hop_en" ] && device_if "$ifname" ch_hop_en "$ch_hop_en"

        config_get ch_long_dur "$vif" ch_long_dur
        [ -n "$ch_long_dur" ] && device_if "$ifname" ch_long_dur "$ch_long_dur"

        config_get ch_nhop_dur "$vif" ch_nhop_dur
        [ -n "$ch_nhop_dur" ] && device_if "$ifname" ch_nhop_dur "$ch_nhop_dur"

        config_get ch_cntwn_dur "$vif" ch_cntwn_dur
        [ -n "$ch_cntwn_dur" ] && device_if "$ifname" ch_cntwn_dur "$ch_cntwn_dur"

        config_get ch_noise_th "$vif" ch_noise_th
        [ -n "$ch_noise_th" ] && device_if "$ifname" ch_noise_th "$ch_noise_th"

        config_get ch_cnt_th "$vif" ch_cnt_th
        [ -n "$ch_cnt_th" ] && device_if "$ifname" ch_cnt_th "$ch_cnt_th"

        config_get_bool scanchevent "$vif" scanchevent
        [ -n "$scanchevent" ] && device_if "$ifname" scanchevent "$scanchevent"

        config_get_bool send_add_ies "$vif" send_add_ies
        [ -n "$send_add_ies" ] && device_if "$ifname" send_add_ies "$send_add_ies"

        config_get_bool ext_ifu_acs "$vif" ext_ifu_acs
        [ -n "$ext_ifu_acs" ] && device_if "$ifname" ext_ifu_acs "$ext_ifu_acs"

        config_get_bool enable_rtt "$vif" enable_rtt
        [ -n "$enable_rtt" ] && device_if "$ifname" enable_rtt "$enable_rtt"

        config_get_bool enable_lci "$vif" enable_lci
        [ -n "$enable_lci" ] && device_if "$ifname" enable_lci "$enable_lci"

        config_get_bool enable_lcr "$vif" enable_lcr
        [ -n "$enable_lcr" ] && device_if "$ifname" enable_lcr "$enable_lcr"

        config_get_bool rrm "$vif" rrm
        [ -n "$rrm" ] && device_if "$ifname" rrm "$rrm"

        config_get_bool rrmslwin "$vif" rrmslwin
        [ -n "$rrmslwin" ] && device_if "$ifname" rrmslwin "$rrmslwin"

        config_get_bool rrmstats "$vif" rrmsstats
        [ -n "$rrmstats" ] && device_if "$ifname" rrmstats "$rrmstats"

        config_get rrmdbg "$vif" rrmdbg
        [ -n "$rrmdbg" ] && device_if "$ifname" rrmdbg "$rrmdbg"

        config_get acparams "$vif" acparams
        [ -n "$acparams" ] && device_if "$ifname" acparams $acparams

        config_get setwmmparams "$vif" setwmmparams
        [ -n "$setwmmparams" ] && device_if "$ifname" setwmmparams $setwmmparams

        config_get_bool qbssload "$vif" qbssload
        [ -n "$qbssload" ] && device_if "$ifname" qbssload "$qbssload"

        config_get_bool proxyarp "$vif" proxyarp
        [ -n "$proxyarp" ] && device_if "$ifname" proxyarp "$proxyarp"

        config_get_bool dgaf_disable "$vif" dgaf_disable
        [ -n "$dgaf_disable" ] && device_if "$ifname" dgaf_disable "$dgaf_disable"

        config_get setibssdfsparam "$vif" setibssdfsparam
        [ -n "$setibssdfsparam" ] && device_if "$ifname" setibssdfsparam "$setibssdfsparam"

        config_get startibssrssimon "$vif" startibssrssimon
        [ -n "$startibssrssimon" ] && device_if "$ifname" startibssrssimon "$startibssrssimon"

        config_get setibssrssihyst "$vif" setibssrssihyst
        [ -n "$setibssrssihyst" ] && device_if "$ifname" setibssrssihyst "$setibssrssihyst"

        config_get son_event_bcast qcawifi son_event_bcast
        [ -n "$son_event_bcast" ] && device_if "$ifname" son_event_bcast "$son_event_bcast"

        config_get noIBSSCreate "$vif" noIBSSCreate
        [ -n "$noIBSSCreate" ] && device_if "$ifname" noIBSSCreate "$noIBSSCreate"

        config_get setibssrssiclass "$vif" setibssrssiclass
        [ -n "$setibssrssiclass" ] && device_if "$ifname" setibssrssiclass $setibssrssiclass

        config_get offchan_tx_test "$vif" offchan_tx_test
        [ -n "$offchan_tx_test" ] && device_if "$ifname" offchan_tx_test $offchan_tx_test

        handle_vow_dbg_cfg() {
                local value="$1"
                [ -n "$value" ] && device_if "$ifname" vow_dbg_cfg $value
        }

        config_list_foreach "$vif" vow_dbg_cfg handle_vow_dbg_cfg

        config_get_bool vow_dbg "$vif" vow_dbg
        [ -n "$vow_dbg" ] && device_if "$ifname" vow_dbg "$vow_dbg"

        handle_set_max_rate() {
                local value="$1"
		if [ "$default_cfg80211" == "1" ]; then
			[ -n "$value" ] && wlanconfig "$ifname" set_max_rate $value -cfg80211
		else
	                [ -n "$value" ] && wlanconfig "$ifname" set_max_rate $value 
		fi
        }
        config_list_foreach "$vif" set_max_rate handle_set_max_rate

        config_get_bool implicitbf "$vif" implicitbf
        [ -n "$implicitbf" ] && device_if "$ifname" implicitbf "${implicitbf}"

        config_get_bool vhtsubfee "$vif" vhtsubfee
        [ -n "$vhtsubfee" ] && device_if "$ifname" vhtsubfee "${vhtsubfee}"

        config_get_bool vhtmubfee "$vif" vhtmubfee
        [ -n "$vhtmubfee" ] && device_if "$ifname" vhtmubfee "${vhtmubfee}"

        config_get_bool vhtsubfer "$vif" vhtsubfer
        [ -n "$vhtsubfer" ] && device_if "$ifname" vhtsubfer "${vhtsubfer}"

        config_get_bool vhtmubfer "$vif" vhtmubfer
        [ -n "$vhtmubfer" ] && device_if "$ifname" vhtmubfer "${vhtmubfer}"

        config_get vhtstscap "$vif" vhtstscap
        [ -n "$vhtstscap" ] && device_if "$ifname" vhtstscap "${vhtstscap}"

        config_get vhtsounddim "$vif" vhtsounddim
        [ -n "$vhtsounddim" ] && device_if "$ifname" vhtsounddim "${vhtsounddim}"

        config_get he_subfee "$vif" he_subfee
        [ -n "$he_subfee" ] && device_if "$ifname" he_subfee "${he_subfee}"

        config_get he_subfer "$vif" he_subfer
        [ -n "$he_subfer" ] && device_if "$ifname" he_subfer "${he_subfer}"

        config_get he_mubfee "$vif" he_mubfee
        [ -n "$he_mubfee" ] && device_if "$ifname" he_mubfee "${he_mubfee}"

        config_get he_mubfer "$vif" he_mubfer
        [ -n "$he_mubfer" ] && device_if "$ifname" he_mubfer "${he_mubfer}"

        config_get he_dlofdma "$vif" he_dlofdma
        [ -n "$he_dlofdma" ] && device_if "$ifname" he_dlofdma "${he_dlofdma}"

        config_get he_ulmumimo "$vif" he_ulmumimo
        [ -n "$he_ulmumimo" ] && device_if "$ifname" he_ulmumimo "${he_ulmumimo}"

        config_get he_dcm "$vif" he_dcm
        [ -n "$he_dcm" ] && device_if "$ifname" he_dcm "${he_dcm}"

        config_get he_extrange "$vif" he_extrange
        [ -n "$he_extrange" ] && device_if "$ifname" he_extrange "${he_extrange}"

        config_get he_ltf "$vif" he_ltf
        [ -n "$he_ltf" ] && device_if "$ifname" he_ltf "${he_ltf}"

        config_get he_txmcsmap "$vif" he_txmcsmap
        [ -n "$he_txmcsmap" ] && device_if "$ifname" he_txmcsmap "${he_txmcsmap}"

        config_get he_rxmcsmap "$vif" he_rxmcsmap
        [ -n "$he_rxmcsmap" ] && device_if "$ifname" he_rxmcsmap "${he_rxmcsmap}"

        config_get ba_bufsize "$vif" ba_bufsize
        [ -n "$ba_bufsize" ] && device_if "$ifname" ba_bufsize "${ba_bufsize}"

        config_get encap_type "$vif" encap_type
        [ -n "$encap_type" ] && device_if "$ifname" encap_type "${encap_type}"

        config_get decap_type "$vif" decap_type
        [ -n "$decap_type" ] && device_if "$ifname" decap_type "${decap_type}"

        config_get_bool rawsim_txagr "$vif" rawsim_txagr
        [ -n "$rawsim_txagr" ] && device_if "$ifname" rawsim_txagr "${rawsim_txagr}"

        config_get clr_rawsim_stats "$vif" clr_rawsim_stats
        [ -n "$clr_rawsim_stats" ] && device_if "$ifname" clr_rawsim_stats "${clr_rawsim_stats}"

        config_get_bool rawsim_debug "$vif" rawsim_debug
        [ -n "$rawsim_debug" ] && device_if "$ifname" rawsim_debug "${rawsim_debug}"

        config_get set_monrxfilter "$vif" set_monrxfilter
        [ -n "$set_monrxfilter" ] && device_if "$ifname" set_monrxfilter "${set_monrxfilter}"

        config_get neighbourfilter "$vif" neighbourfilter
        [ -n "$neighbourfilter" ] && device_if "$ifname" neighbourfilter "${neighbourfilter}"

        config_get osen "$vif" osen
        [ -n "$osen" ] && device_if "$ifname" osen "$osen"

        if [ $osen -ne 0 ]; then
                device_if "$ifname" proxyarp 1
        fi

        config_get re_scalingfactor "$vif" re_scalingfactor
        [ -n "$re_scalingfactor" ] && device_if "$ifname" set_whc_sfactor "$re_scalingfactor"

        config_get ul_hyst "$vif" ul_hyst
        [ -n "$ul_hyst" ] && device_if "$ifname" ul_hyst "${ul_hyst}"

        config_get root_distance "$vif" root_distance
        [ -n "$root_distance" ] && device_if "$ifname" set_whc_dist "$root_distance"

        config_get caprssi "$vif" caprssi
        [ -n "$caprssi" ] && device_if "$ifname" caprssi "${caprssi}"

        config_get_bool ap_isolation_enabled $device ap_isolation_enabled 0
        config_get_bool isolate "$vif" isolate 0

        if [ $ap_isolation_enabled -ne 0 ]; then
                [ "$mode" = "wrap" ] && isolate=1
                device_if "$phy" isolation "$ap_isolation_enabled"
        fi

        config_get_bool ctsprt_dtmbcn "$vif" ctsprt_dtmbcn
        [ -n "$ctsprt_dtmbcn" ] && device_if "$ifname" ctsprt_dtmbcn "${ctsprt_dtmbcn}"

        config_get assocwar160  "$vif" assocwar160
        [ -n "$assocwar160" ] && device_if "$ifname" assocwar160 "$assocwar160"

        config_get rawdwepind "$vif" rawdwepind
        [ -n "$rawdwepind" ] && device_if "$ifname" rawdwepind "$rawdwepind"

        config_get revsig160  "$vif" revsig160
        [ -n "$revsig160" ] && device_if "$ifname" revsig160 "$revsig160"

        config_get channel_block_list "$vif" channel_block_list
        [ -n "$channel_block_list" ] && wifitool "$ifname" block_acs_channel "$channel_block_list"

        config_get rept_spl  "$vif" rept_spl
        [ -n "$rept_spl" ] && device_if "$ifname" rept_spl "$rept_spl"

        config_get cactimeout  "$vif" cactimeout
        [ -n "$cactimeout" ] && device_if "$ifname" set_cactimeout "$cactimeout"

        config_get meshdbg "$vif" meshdbg
        [ -n "$meshdbg" ] && device_if "$ifname" meshdbg "$meshdbg"

        config_get rmode_pktsim "$vif" rmode_pktsim
        [ -n "$rmode_pktsim" ] && device_if "$ifname" rmode_pktsim "$rmode_pktsim"

        config_get global_wds qcawifi global_wds

        if [ $global_wds -ne 0 ]; then
                device_if "$ifname" athnewind 1
        fi

        config_get pref_uplink "$device" pref_uplink
        [ -n "$pref_uplink" ] && device_if "$phy" pref_uplink "${pref_uplink}"

        config_get fast_lane "$device" fast_lane
        [ -n "$fast_lane" ] && device_if "$phy" fast_lane "${fast_lane}"

        if [ $fast_lane -ne 0 ]; then
                device_if "$ifname" athnewind 1
        fi

}

common_vap_setting_after_vap_up() {

	vif=$1
	config_get ifname "$vif" ifname	

	config_get_bool vht_11ng "$vif" vht_11ng
        [ -n "$vht_11ng" ] && device_if "$ifname" vht_11ng "$vht_11ng"

        config_get vhtmcs "$vif" vhtmcs
        [ -n "$vhtmcs" ] && device_if "$ifname" vhtmcs "$vhtmcs"

        config_get dis_legacy "$vif" dis_legacy
        [ -n "$dis_legacy" ] && device_if "$ifname" dis_legacy "$dis_legacy"

        config_get mbo "$vif" mbo
        [ -n "$mbo" ] && device_if "$ifname" mbo "$mbo"

        if [ $mode = "sta" ]; then
                config_get enable_ft  "$vif" ieee80211r
                [ -n "$enable_ft" ] && device_if "$ifname" ft "$enable_ft"
        fi

        config_get enable_fils  "$vif" ieee80211ai
        config_get fils_discovery_period  "$vif" fils_fd_period 0
        [ -n "$enable_fils" ] && device_if "$ifname" enable_fils "$enable_fils" "$fils_discovery_period"

        config_get bpr_enable  "$vif" bpr_enable
        [ -n "$bpr_enable" ] && device_if "$ifname" set_bpr_enable "$bpr_enable"

        config_get oce "$vif" oce
        [ -n "$oce" ] && device_if "$ifname" oce "$oce"
        [ "$oce" -gt 0 ] && {
        	case "$hwmode" in
	                11B*|11G*|11NG*)
	                        device_if "$ifname" set_bcn_rate 5500
        	                device_if "$ifname" prb_rate 5500
                	        ;;
			*)
                        	;;
                esac
                [ -z "$enable_fils" ] && {
                	config_get fils_discovery_period  "$vif" fils_fd_period 20
	                device_if "$ifname" enable_fils 1 "$fils_discovery_period"
                }
	}

        config_get set_bcn_rate "$vif" set_bcn_rate
        [ -n "$set_bcn_rate" ] && device_if "$ifname" set_bcn_rate "$set_bcn_rate"

        config_get mcast_rate "$vif" mcast_rate
        [ -n "$mcast_rate" ] && device_if "$ifname" mcast_rate "${mcast_rate%%.*}"
        
	# Enable RPS and disable qrfs, if rxchainmask is 15 for some platforms
        enable_rps_wifi=0
	config_get device "$vif" device
        if [ $(device_if "$device" get_rxchainmask | awk -F ':' '{ print $2 }') -gt 3 ]; then
                enable_rps_wifi=1
        fi
        if [ $enable_rps_wifi == 1 ] && [ -f "/lib/update_system_params.sh" ]; then
                . /lib/update_system_params.sh
                enable_rps $ifname
        fi

}

common_dev_setting_after_vap_up() {

	wifi_dev=$1
	
        # Enable RPS and disable qrfs, if rxchainmask is 15 for some platforms
        disable_qrfs_wifi=0
        if [ $(device_if "$phy" get_rxchainmask | awk -F ':' '{ print $2 }') -gt 3 ]; then
                disable_qrfs_wifi=1
        fi

	config_get wifi_debug_sh $device wifi_debug_sh
	[ -n "$wifi_debug_sh" -a -e "$wifi_debug_sh" ] && sh "$wifi_debug_sh"

        config_get primaryradio "$wifi_dev" primaryradio
        [ -n "$primaryradio" ] && device_if "$wifi_dev" primaryradio "${primaryradio}"

        config_get CSwOpts "$wifi_dev" CSwOpts
        [ -n "$CSwOpts" ] && device_if "$wifi_dev" CSwOpts "${CSwOpts}"

        if [ $disable_qrfs_wifi == 1 ] && [ -f "/lib/update_system_params.sh" ]; then
                . /lib/update_system_params.sh
                disable_qrfs
        fi

	#move below flow to load_qcawifi, from SPF10.0
#        if [ -f "/lib/update_smp_affinity.sh" ]; then
#                config_load wireless
#                . /lib/update_smp_affinity.sh
#                config_foreach enable_smp_affinity_wifi wifi-device
#        fi
	device=$wifi_dev
}

set_hostapd_wpa_supplicant() {

		vif=$1

		config_get ifname "$vif" ifname
		config_get mode "$vif" mode
		config_get enc "$vif" encryption "none"
		config_get nawds "$vif" nawds

                func_enable_radius_acc=$(uci -q get functionlist.functionlist.SUPPORT_ALL_ENCRYPTION_HAS_RADIUS_SETTING)

                if [ "$func_enable_radius_acc" == "1" ]; then
                        config_get acct_enabled "$vif" acct_enabled
                        if [ "$acct_enabled" == "1" ]; then
                                # Radius Accounting enable for Disabled/WEP/PSK*
                                start_hostapd=1
                        fi
                fi

                [ -f /var/run/hostapd-$ifname.conf ] && rm -f /var/run/hostapd-$ifname.conf

		config_get own_ie_override "$vif" own_ie_override
		[ -n "$own_ie_override" ] && device_if "$ifname" rsn_override 1	#jacky: in qcawificfg80211.sh, it uses device_if here

		config_get_bool sae "$vif" sae
		config_get_bool owe "$vif" owe
		config_get_bool dpp "$vif" dpp
		config_get pkex_code "$vif" pkex_code
		config_get suite_b "$vif" suite_b 0

		if [ $suite_b -eq 192 ]
		then
			cat /sys/class/net/$device/ciphercaps | grep -i "gcmp-256"
			if [ $? -ne 0 ]
			then
				echo "enc:GCMP-256 is Not Supported on Radio" > /dev/console
				destroy_vap $ifname
				continue
			fi
		elif [ $suite_b -ne 0 ]
		then
			echo "$suite_b bit security level is not supported for SUITE-B" > /dev/console
			destroy_vap $ifname
			continue
		fi

		if [ "${dpp}" -eq 1 ]
		then
			device_if "$ifname" set_dpp_mode 1
			config_get dpp_type "$vif" dpp_type "qrcode"
			if [ "$dpp_type" != "qrcode" -a "$dpp_type" != "pkex" ]
			then
				echo "Invalid DPP type" > /dev/console
				destroy_vap $ifname
				continue
			elif [ "$dpp_type" == "pkex" ]
			then
				if [ -z "$pkex_code" ]
				then
					echo "pkex_code should not be NULL" > /dev/console
					destroy_vap $ifname
				fi
			fi
		fi

		check_sae_groups() {
			local sae_groups=$(echo $1 | tr "," " ")
			for sae_group_value in $sae_groups
			do
				case "$sae_group_value" in
					15|16|17|18|19|20|21|28|29|30)
					;;
					*)
						echo "Invalid sae_group: $sae_group_value" > /dev/console
						destroy_vap $ifname
						is_valid_sae_group=1
						break
				esac
			done
		}


                case "$enc" in
                        none)
                                # If we're in open mode and want to use WPS, we
                                # must start hostapd
                                config_get_bool wps_pbc "$vif" wps_pbc 0
                                config_get config_methods "$vif" wps_config
                                [ "$wps_pbc" -gt 0 ] && append config_methods push_button
                                start_wpa_supplicant=1 # Dakota need wpa_supplicant at none encrypt
                                start_hostapd=1
                                # We start hostapd in open mode also
                        ;;

                        wep*)
                                config_get key "$vif" key_id
                                config_get wep_key "$vif" key$key
                                [ "$mode" = "sta" ] || {
                                        iwconfig "$ifname" enc "[$key]" "${wep_key:-off}"
                                        case "$enc" in
                                                *open)
                                                        authmode=1;authmode2="open"
                                                        ;;
                                                *shared)
                                                        authmode=2;authmode2="restricted"
                                                        ;;
                                        esac
                                        device_if "$ifname" authmode $authmode
                                        iwconfig "$ifname" key $authmode2 # restricted,open
                                }
                                start_hostapd=1
				
				if [ $sae -eq 1 ] || [ $owe -eq 1 ]
				then
					echo "With SAE/OWE enabled, wep enc is not supported" > /dev/console
					destroy_vap $ifname
					continue
				fi
                        ;;
			wpa*|8021x|osen*)
                                start_hostapd=1
                                start_wpa_supplicant=1
                                config_get key "$vif" key

                                ### SENAO ### 11k fastroaming
				#config_get fastroaming_en "$vif" fastroamingEnable
                                #if [ "$fastroaming_en" == "1" ]; then
                                #        device_if "$ifname" rrm 1
                                #else   #no need to set zero since it already preset before
                                #       device_if "$ifname" rrm 0
                                #fi
				
				start_hostapd=1
			;;
			mixed*|psk*)
                                start_hostapd=1
                                start_wpa_supplicant=1
                                config_get key "$vif" key

                                ### SENAO ### 11k fastroaming
				#config_get fastroaming_en "$vif" fastroamingEnable
                                #if [ "$fastroaming_en" == "1" ]; then
                                #        device_if "$ifname" rrm 1
                                #else   #no need to set zero since it already preset before
                                #       device_if "$ifname" rrm 0
                                #fi

				config_get wpa_psk_file  "$vif" wpa_psk_file
				if [ -z "$key" ] && [ -z "$wpa_psk_file" ]
                then
					echo "Key is NULL" > /dev/console
					destroy_vap $ifname
					continue
				fi
	
				case "$enc" in
					*tkip*)
						if [ $sae -eq 1 ] || [ $owe -eq 1 ]
						then
							echo "With SAE/OWE enabled, tkip enc is not supported" > /dev/console
							destroy_vap $ifname
							continue
						fi
					;;
				esac

				if [ $sae -eq 1 ]
				then
					config_get sae_password "$vif" sae_password

					if [ -z "$sae_password" ] && [ -z "$key" ] && [ -z "$wpa_psk_file" ]
					then
						echo "key and sae_password are NULL" > /dev/console
						destroy_vap $ifname
						continue
					fi

					case "$sae_password" in
						*|id=*)
						echo "SAE PW + PWID supported"
						device_if "$ifname" en_sae_pwid 1
						;;
					esac

					case "$key" in
						*|id=*)
						echo "SAE PW + PWID supported"
						device_if "$ifname" en_sae_pwid 1
						;;
					esac
					config_list_foreach "$vif" sae_groups check_sae_groups

					[ $is_valid_sae_group = 1 ] && continue
				fi
			;;
			tkip*)
				if [ $sae -eq 1 ] || [ $owe -eq 1 ]
				then
					echo "With SAE/OWE enabled, tkip enc is not supported" > /dev/console
					destroy_vap $ifname
					continue
				fi
                        ;;
			#Needed ccmp*|gcmp* check for SAE OWE auth types
			ccmp*|gcmp*)
				start_hostapd=1
				start_wpa_supplicant=1
				config_get key "$vif" key
				config_get wpa_psk_file  "$vif" wpa_psk_file
				pwid_exclusive=1
				pwid=0
				sae_password_count=0
				if [ $sae -eq 1 ]
				then
					check_sae_pwid_capab() {
						local sae_password=$(echo $1)
						if [ -n "$sae_password" ]
						then
							sae_password_count=$((sae_password_count+1))

							case "$sae_password" in
								*"|id="*)
								pwid=1
								echo "pwid match" > /dev/console
								;;
								*)
								echo "pw only match" > /dev/console
								pwid_exclusive=0
								;;
							esac
						fi
					}
					config_list_foreach "$vif" sae_password check_sae_pwid_capab
					echo "sae password count:" $sae_password_count > /dev/console

					if [ sae_password_count -eq 0] && [ -z "$key" ] && [ -z "$wpa_psk_file" ]
					then
						echo "key and sae_password are NULL" > /dev/console
						destroy_vap $ifname
						continue
					fi
					if [ $pwid -eq 1 ]
					then
						if [ $pwid_exclusive -eq 1 ]
						then
							device_if "$ifname" en_sae_pwid 3
						else
							device_if "$ifname" en_sae_pwid 1
						fi
					fi
					config_list_foreach "$vif" sae_groups check_sae_groups

					[ $is_valid_sae_group = 1 ] && continue
				fi
				if [ $owe -eq 1 ]
				then
					if [ "$mode" = "ap" ]
					then
						check_owe_groups() {
							local owe_groups=$(echo $1 | tr "," " ")
							for owe_group_value in $owe_groups
							do
								case "$owe_group_value" in
									19|20|21)
									;;
									*)
										echo "Invalid owe_group: $owe_group_value" > /dev/console
										destroy_vap $ifname
										is_valid_owe_group=1
										break
								esac
							done
						}
						config_list_foreach "$vif" owe_groups check_owe_groups
					elif [ "$mode" = "sta" ]
					then
						config_get owe_group "$vif" owe_group 0
						case "$owe_group" in
							0|19|20|21)
							;;
							*)
								echo "Invalid owe_group: $owe_group" > /dev/console
								destroy_vap $ifname
								is_valid_owe_group=1
								break
						esac
					fi

					[ $is_valid_owe_group = 1 ] && continue
				fi
			;;
			#added below from spf10.0
			sae*|dpp|psk2)
				start_hostapd=1
			;;
                esac

		#jacky: WAR for hostapd abnormal issue, add sleep to avoid bring hostapd too frequently
		sleep 1

		if [ "$default_cfg80211" == "1" ]; then
			hostapd_drv="nl80211"
			wpa_supplicant_drv="nl80211"
		else
			hostapd_drv="atheros"
			wpa_supplicant_drv="athr"
		fi

                case "$mode" in
                        ap|wrap|ap_monitor|ap_smart_monitor|mesh|ap_lp_iot)
                                device_if "$ifname" ap_bridge "$((isolate^1))"
                                if [ -n "$start_hostapd" ] && [ "$default_cfg80211" == "1" -o "$nawds" != "1" ] && eval "type hostapd_setup_vif" 2>/dev/null >/dev/null; then
                                        hostapd_setup_vif "$vif" "$hostapd_drv" no_nconfig || {
                                                DebugPrint "enable_qcawifi($device): Failed to set up hostapd for interface $ifname"
                                                # make sure this wifi interface won't accidentally stay open without encryption
                                                ifconfig "$ifname" down
						if [ "$default_cfg80211" == "1" ]; then
							iw $ifname del 
						else
                                                	wlanconfig "$ifname" destroy
						fi
                                                continue
                                        }
                                fi
                        ;;
                        wds|sta)
                                # Client Bridge + WPA* = call wpa_supplicant_setup_vif
                                # Client Bridge + WEP = don't call wpa_supplicant_setup_vif (not sure at Dakota)
                                # Client Bridge + None + Dakota= call wpa_supplicant_setup_vif
                                # Client Bridge + None + Scorpion= don't wpa_supplicant_setup_vif

                                # Repeater + WPA*/WEP/None = call wpa_supplicant_setup_vif
                                if [ "$mode_display" != "sta" -o "$start_wpa_supplicant" = "1" ] && eval "type wpa_supplicant_setup_vif" 2>/dev/null >/dev/null; then
                                        wpa_supplicant_setup_vif "$vif" "$wpa_supplicant_drv" || {
                                                DebugPrint "enable_qcawifi($device): Failed to set up wpa_supplicant for interface $ifname"
                                                ifconfig "$ifname" down
						if [ "$default_cfg80211" == "1" ]; then
                                                        iw $ifname del
                                                else
	                                                wlanconfig "$ifname" destroy
						fi
                                                continue
                                        }
                                fi
                        ;;
                        adhoc)
                                if eval "type wpa_supplicant_setup_vif" 2>/dev/null >/dev/null; then
                                        wpa_supplicant_setup_vif "$vif" "$wpa_supplicant_drv" || {
                                                DebugPrint "enable_qcawifi($device): Failed to set up wpa"
                                                ifconfig "$ifname" down
						if [ "$default_cfg80211" == "1" ]; then
                                                        iw $ifname del
                                                else
	                                                wlanconfig "$ifname" destroy
						fi
                                                continue
                                        }
                                fi
                        ;;
                esac

}

set_hwmode_htmode() {
		local ifname=$1
		local device=$2
		local opmode=$3
                config_get hwmode "$device" hwmode auto
                config_get htmode "$device" htmode auto
		config_get edge_channel_deprioritize "$device" edge_channel_deprioritize 1
		config_get channel "$device" channel 0
                local pureg=0
                local puren=0
                local ac_mode=0
                local disablecoext=0 # we can use this parameter to set fixed40Mhz
                local chwidth=0
                local allow_frag=0

		 [ "auto" == "$channel" ] && channel=0
		# WAR to not use chan 36 as primary channel, when using higher BW.
		if [ $channel -eq 36 ]; then
			if [ $edge_channel_deprioritize -eq 1 ]; then
				case "$board_name" in
					ap-hk*|ap-ac*|ap-oa*)
					[ 36 = "$channel" ] && [ "HT20" != "$htmode" ] && channel=40 &&  DebugPrint " Primary channel is changed to 40"
					[ "HT40+" = "$htmode" ] && htmode="HT40-" && DebugPrint " Mode changed to HT40MINUS with channel 40"
				;;
					*)
				;;
				esac
			fi
		fi

                # if it is client bridge, give hwmode and htmode with 11ng and HT40 for BGN mode, and 11ac and HT80 for 11ACVHT80 mode.
                # let sta connect to ap in 20MHz or 40MHz bandwidth which depend on AP's setting.
                if [ "$opmode" == "sta" -o "$opmode" == "wds_sta" -o "$opmode" == "sta_ap" ]; then
                        channel=0
                        config_set "$device" aggregation_enable "0"
			hwmodePath="/sys/class/net/$device/hwmodes"
			if [ -e $hwmodePath ]; then	#new using for cb mode choose max speed hwmode
				if [ "$(grep -c "11AXA" $hwmodePath)" != "0" ]; then
					hwmode=11axa; htmode=HT80
					config_set "$device" rate ""
				elif [ "$(grep -c "11AXG" $hwmodePath)" != "0" ]; then
					hwmode=11axg; htmode=HT20_40
                                        config_set "$device" rate "0x0"
				elif [ "$(grep -c "11AC" $hwmodePath)" != "0" ]; then
					hwmode=11ac; htmode=HT80                                
                                        config_set "$device" rate ""
				else
					hwmode=11ng; htmode=HT20_40
                                        config_set "$device" rate "0x0"	
				fi
			else
        	                case "$device" in
        	                        *0)
        	                                hwmode=11ng; htmode=HT20_40
        	                                config_set "$device" rate "0x0"
        	                                ;;
        	                        *1|*2)
        	                                hwmode=11ac; htmode=HT80                                        
						config_set "$device" rate ""
        	                        ;;
	                        esac
			fi
                fi

                [ -f /tmp/opmode ] || touch /tmp/opmode
                uci -c /tmp -q get opmode.@$device[0] || {
                        uci -c /tmp -q add opmode $device
                        uci -c /tmp -q set opmode.@$device[0].opmode=$opmode
                }

                # if not ap mode or opmode change, renew dhcp
                [ "$renew" = "0" -a "$opmode" != "ap" -o "$opmode" != "$(uci -c /tmp get opmode.@$device[0].opmode)" ] && {
                        proto=$(uci get network.lan.proto)
                        ${proto}_renew
                        renew=1
                        unset proto
                }

                # record opmode
                uci -c /tmp -q set opmode.@$device[0].opmode=$opmode
                uci -c /tmp -q commit opmode

                case "$hwmode:$htmode" in
                # The parsing stops at the first match so we need to make sure
                # these are in the right orders (most generic at the end)
                        *ng:HT20) hwmode=11NGHT20; disablecoext=1; chwidth=0;;
                        *ng:HT20_40) hwmode=11NGHT40; chwidth=1;;
                        *ng:HT40) hwmode=11NGHT40; disablecoext=1; chwidth=1;;
                        *ng:*) hwmode=11NGHT20; chwidth=0;;
                        *na:HT20) hwmode=11NAHT20; chwidth=0;;
                        *na:HT20_40) hwmode=11NAHT40; chwidth=1;;
                        *na:HT40) hwmode=11NAHT40; disablecoext=1; chwidth=1;;
                        *na:*) hwmode=11NAHT40; chwidth=1;;
                        *ac:HT20) hwmode=11ACVHT20; ac_mode=1; chwidth=0;;
                        *ac:HT40+) hwmode=11ACVHT40PLUS; ac_mode=1; chwidth=1;;
                        *ac:HT40-) hwmode=11ACVHT40MINUS ac_mode=1; chwidth=1;;
                        *ac:HT20_40) hwmode=11ACVHT40; ac_mode=1; chwidth=1;;
                        *ac:HT40) hwmode=11ACVHT40; ac_mode=1; disablecoext=1; chwidth=1;;
                        *ac:HT80) hwmode=11ACVHT80; ac_mode=1; chwidth=2;;
                        *ac:HT160) hwmode=11ACVHT160; ac_mode=1; chwidth=3;;
                        *ac:HT80_80) hwmode=11ACVHT80_80; chwidth=2;;
			*axg:HT20) hwmode=11GHE20; disablecoext=1; chwidth=0;;
			*axg:HT40-) hwmode=11GHE40MINUS; disablecoext=1; chwidth=1;;
			*axg:HT40+) hwmode=11GHE40PLUS; disablecoext=1; chwidth=1;;
			*axg:HT20_40) hwmode=11GHE40; chwidth=1;;
			*axg:HT40) hwmode=11GHE40; disablecoext=1; chwidth=1;;
			*axg:*) hwmode=11GHE20; chwidth=0;;
			*axa:HT20) hwmode=11AHE20;  disablecoext=1; chwidth=0;;
			*axa:HT40+) hwmode=11AHE40PLUS; disablecoext=1; chwidth=1;;
			*axa:HT40-) hwmode=11AHE40MINUS; disablecoext=1; chwidth=1;;
			*axa:HT20_40) hwmode=11AHE40; chwidth=1;;
			*axa:HT40) hwmode=11AHE40; disablecoext=1; chwidth=1;;	
			*axa:HT80) hwmode=11AHE80; disablecoext=1; chwidth=2;;
			*axa:HT160) hwmode=11AHE160; chwidth=3;;
			*axa:HT80_80) hwmode=11AHE80_80; chwidth=3;;
			*axa:*) hwmode=11AHE80
				if [ -f /sys/class/net/$device/5g_maxchwidth ]; then
					maxchwidth="$(cat /sys/class/net/$device/5g_maxchwidth)"
					[ -n "$maxchwidth" ] && hwmode=11AHE$maxchwidth
				fi
				chwidth=2
				if [ "$mode" == "sta" ]; then
					cat /sys/class/net/$device/hwmodes | grep  "11AXA_HE80_80"
					if [ $? -eq 0 ]; then
						hwmode=11AHE80_80
						chwidth=3
					fi
				fi;;
                        *ac:*) hwmode=11ACVHT80; ac_mode=1; chwidth=2;;
                        *n:*) puren=1;;
                        *b:*) hwmode=11B; allow_frag=1;;
                        *bg:*) hwmode=11G; allow_frag=1;;
                        *g:*) hwmode=11G; pureg=1; allow_frag=1;;
                        *a:*) hwmode=11A; allow_frag=1;;
                        *) hwmode=AUTO;;
                esac

                #handle puren 
                [ $puren -gt 0 ] && {
                        case "$device:$htmode" in
                                *0:HT20) hwmode=11NGHT20; disablecoext=1; chwidth=0;;
                                *0:HT20_40) hwmode=11NGHT40; chwidth=1;;                                
                                *0:HT40) hwmode=11NGHT40; disablecoext=1; chwidth=1;;
                                *:HT20) hwmode=11NAHT20; chwidth=0;;
                                *:HT20_40) hwmode=11NAHT40; chwidth=1;;
                                *:HT40) hwmode=11NAHT40; disablecoext=1; chwidth=1;;
                                *) hwmode=AUTO;;
                        esac
                }
                device_if "$ifname" mode "$hwmode"
                [ $pureg -gt 0 ] && device_if "$ifname" pureg "$pureg"
                [ $puren -gt 0 ] && device_if "$ifname" puren "$puren"
                
## pre-CAC
                ## set zero-wait DFS result to working radio
                set_dfs_nol_list "$device" "$opmode"
## pre-CAC
      
		#cfg80211 auto channel issue, we need to assign a default channel to bring up interface
                [ "$default_cfg80211" == "1" ] && [ "$channel" == "0" ] && [ "$mode" != "monitor" ] && device_if "$ifname" channel 0
                                                                        
                [ $channel -gt 0 ] && [ "$mode" != "monitor" ] && iwconfig "$ifname" channel "$channel"

                device_if "$ifname" disablecoext "${disablecoext}"
                device_if "$ifname" chwidth "${chwidth}"

                config_get frag "$vif" frag
                config_get rts "$device" rts

		if [ "$default_cfg80211" == "1" ]; then
			# configure below options once AP is up , as wireless mode is available now
			#[ -n "$frag" ] && iw phy "$(cat /sys/class/net/$phy/phy80211/name)"  set frag "${frag%%.*}"
			[ -n "$rts" ] && iw phy "$(cat /sys/class/net/$device/phy80211/name)" set rts "${rts%%.*}"
		else
	                if [ -n "$rts" ] && [ "$ac_mode" != "1" ]; then
        	                [ ${rts} -ge 2346 ] && unset rts
				if [ "$default_cfg80211" == "1" ]; then
		                        # configure below options once AP is up , as wireless mode is available now
					[ -n "$rts" ] && iw phy "$(cat /sys/class/net/$device/phy80211/name)" set rts "${rts%%.*}"
				else
                		        [ -n "$rts" ] && iwconfig "$ifname" rts "${rts%%.*}"
				fi
			
	                fi
			if [ "$default_cfg80211" == "1" ]; then
				[ -n "$rts" -a "$ac_mode" == "1" ] && iw phy "$(cat /sys/class/net/$device/phy80211/name)" set rts off
			else
	        	        [ -n "$rts" -a "$ac_mode" == "1" ] && iwconfig "$ifname" rts off
			fi
		fi

}

set_nawds_or_mesh_mode() {
	local wifi_dev=$1
	config_get vifs "$wifi_dev" vifs

	[ -z $vifs ] && vifs=$(eval "/usr/sbin/foreach wireless wifi-iface device $wifi_dev")

	for vif in $vifs; do
                config_get ifname "$vif" ifname
                config_get nawds "$vif" nawds
                config_get disabled "$vif" disabled
                config_get opmode "$wifi_dev" opmode
                config_get mode_display "$vif" mode_display
                config_get mesh "$vif" mesh 0
                [ "$mesh" != "1" ] && [ "$opmode" != "$mode_display" ] && {
                        continue
                }

                local mesh_disabled=$(uci -q get mesh.wifi.disabled)

                [ "$mesh" == "1" ] && [ "$mesh_disabled" == "1" ] && continue;

                case "$nawds:$disabled:$mesh" in
                1:0:1)
		        support_mesh=$(get_sn_wifi_option SUPPORT_NAWDS_MESH)

        		[ "$support_mesh" == "1" ] && {
                        
				enable_mesh_nawds   "$ifname" "$vif"
				### mesh easy-setup v1 ###
                        	local MeshEzSetup=$(uci -q get wireless.$vif.MeshEasySetup)

				support_mesh_ezsetup=$(get_sn_wifi_option SUPPORT_MESH_EASYSETUP)

	                        if [ "$MeshEzSetup" == "1" -a "$support_mesh_ezsetup" == "1" ]; then
        	                        enable_mesh_easysetup "$ifname" "$vif"
                	        fi
			}
			### end ###
                ;;
                1:0:0)
                        enable_wdsbridge   "$ifname" "$vif"
			wds_enable=1
                ;;
                esac
		### mesh easy-setup v2 ###
		local t_C_24g=$(uci -q get network.sys.mesh_configured)
		local t_C_5g=$(uci -q get network.sys.mesh_configured_5g)

		[ -z "$t_C_24g" ] && t_C_24g=0
		[ -z "$t_C_5g" ] && t_C_5g=0

		local t_Configured=$(expr $t_C_24g + $t_C_5g)
		local t_EzSetup=$(uci -q get wireless.$vif.MeshEasySetup)

		if [ "$t_EzSetup" = "2" ] && [ "$t_Configured" = "0" ]; then
			enable_mesh_easysetup_cfg "$vif"
		fi
		### end ###
	done
}

sn_fix_rate_issue_func() {
	local wifi_dev=$1
        local rate_ioctl
      	local ac_mode=$(uci -q get wireless.$wifi_dev.hwmode | grep ac)
	
	[ -n "$ac_mode" ] && ac_mode=1

	config_get rate "$wifi_dev" rate
        [ -n "$rate" ] && [ "$ac_mode" == "1" ] && rate_ioctl=vhtmcs || rate_ioctl=set11NRates

        for ifn in $all_ifnames; do
                ifconfig $ifn down
                ifconfig $ifn up
                DebugPrint "enable interface $ifn and use iwconfig to check interface down or up"
                # Fix "Interface doesn't accept private ioctl, set11NRates (8BE0): Invalid argument"
                [ -n "$rate" ] && delay_cmd 5 device_if "$ifn" $rate_ioctl "$rate"
                if [ -n "$rate" ] && [ "$rate" != "0x0" ] && [ "$rate_ioctl" == "set11NRates" ] && [[ "$wifi_dev" == "wifi1" || "$wifi_dev" == "wifi2" ]]; then
                        device_if $ifn ldpc 0x0                    #disable ldpc for IOT issue, only for 5G 11N fixed data rate
                else
                        if [ -n "$ldpc" ]; then
                                device_if "$ifn" ldpc "$ldpc"
                        else
                                device_if "$ifn" ldpc 0x3                  #default setting for ldpc
                        fi
                fi
        done

}

sn_fix_HT20_40_issue_func() {
        # Fix: if HT20_40 cfg, let 2.4g choose proper bandwidth itself when locates in noisy wlan env(mostly it choose 20MHz)
	local wifi_dev=$1
        config_get htmode "$device" htmode auto
        case "$device:$htmode" in
                *0:HT20_40)
                        sleep 3
                        for ifn in $all_ifnames; do
                                (ifconfig $ifn down && sleep 2 && ifconfig $ifn up && DebugPrint "$ifn:down-and-up" )&
                        done
                        ;;
        esac
}

fix_11ax_ap_panic_issue() {
	# This is QCA's bug. Two Samsung cellphones(support 11ax) connect to 11ax 5G ap. When they perform FTP file transfer,
	# The ap causes kernel panic in a short time. The ap not happen in 11ac mode.
	for ifn in $all_ifnames; do
		DebugPrint "wifitool $ifn setUnitTestCmd 0x47 2 29 1"
		wifitool $ifn setUnitTestCmd 0x47 2 29 1
	done
}

sn_set_channel_final() {

        [ "$default_cfg80211" == "1" ] && return # avoid auto chan cause vap not work in non-dfs chan.

        wifi_dev=$1
              
        config_get channel "$wifi_dev" channel                        
        config_get vifs "$wifi_dev" vifs
        config_get opmode "$wifi_dev" opmode

	 [ "auto" == "$channel" ] && channel=0
	
        [ $setchannel -eq 0 -a "$opmode" != "sta" -a "$opmode" != "wds_sta" -a "$opmode" != "sta_ap" ] && {
		if [ -n $lastApIf ]; then
                        delay_cmd 5 iwconfig "$lastApIf" channel "$channel"
                else
                        delay_cmd 5 iwconfig "$ifname" channel "$channel"
		fi
                setchannel=1;
        }
}

set_channel_config() {

    support_channelconfig=$(get_sn_wifi_option SUPPORT_CHANNEL_CONFIG)
    [ "$support_channelconfig" == "0" ] && return
    
    /sbin/setChannelConfig.sh $1

}

disable_scan_radio(){

	wifi_dev=$1

	config_get opmode "$device" opmode

	if [ ! -e "/sbin/ScanRadio.sh" -o "$opmode" != "mon" ]; then
		return 0
	fi

	test "$(pgrep -f ScanRadio)" && kill -9 $(pgrep -f ScanRadio)
	if [ $sniffer_enable -eq 1 ]; then 
		test "$(pgrep -f snsniffer)" && kill -9 $(pgrep -f snsniffer)	#maybe no need kill, this process terminated when interface down
	else
		test "$(pgrep -f airodump-ng)" && kill -9 $(pgrep -f airodump-ng)	#maybe no need kill, this process terminated when interface down
	fi

	rm /tmp/ScanRadio/ -rf	#remove ap sta scan results
	[ -e "/sys/class/leds/scan_led/brightness" ] && echo 0 > /sys/class/leds/scan_led/brightness	#add protection if no scan_led
}

set_scan_radio(){
	
	wifi_dev=$1

        config_get opmode "$device" opmode
        config_get spectrumMode "$device" spectrumMode 0

        [ "$wifi_dev" == "wifi0" -o "$wifi_dev" == "wifi1" ] && [ "$spectrumMode" == "0" ] && return

        [ "$opmode" == "ap" -a "$spectrumMode" == "1" ] && {
                config_get vifs "$wifi_dev" vifs

		[ -z $vifs ] && vifs=$(eval "/usr/sbin/foreach wireless wifi-iface device $wifi_dev")

                for vif in $vifs; do
                        config_get disabled "$vif" disabled
                        config_get mode_display "$vif" mode_display
                        [ "$opmode" == "$mode_display" -a "$disabled" == "0" ] && {
                                scanRadioFolder="/tmp/ScanRadio/"
                                [ ! -d "$scanRadioFolder" ] && mkdir -p "$scanRadioFolder"

                               if [ -e "/sbin/ScanRadio.sh" ]; then
                                       sh /sbin/ScanRadio.sh $vif 10 $wifi_dev &
                               else
                                       DebugPrint "/sbin/ScanRadio.sh not exists"
                               fi
                        }
                done
        }

        [ "$opmode" == "mon" ] && {
                config_get vifs "$wifi_dev" vifs

		[ -z $vifs ] && vifs=$(eval "/usr/sbin/foreach wireless wifi-iface device $wifi_dev")

                for vif in $vifs; do
                        config_get ifname "$vif" ifname
                        config_get mode_display "$vif" mode_display
                        [ "$opmode" == "$mode_display" ] && {

				scanRadioFolder="/tmp/ScanRadio/"
				[ ! -d "$scanRadioFolder" ] && mkdir -p "$scanRadioFolder"
				
				ifconfig $ifname up	#not regular flow, just up it before works
				[ -e "/sys/class/leds/scan_led/brightness" ] && echo 1 > /sys/class/leds/scan_led/brightness

				config_get scanProgress "$vif" scanProgress #script to switch channel
				if [ "$scanProgress" == "1" ]; then
					if [ -e "/sbin/ScanRadio.sh" ]; then
						config_get scanInterval "$vif" scanInterval #if value is null, it sets default by ScanRadio.sh.
						sh /sbin/ScanRadio.sh $vif $scanInterval $wifi_dev &
					else
						DebugPrint "/sbin/ScanRadio.sh not exists"
					fi
				fi
				if [ $sniffer_enable -eq 1 ]; then 
					config_get snsnifferEnable "$vif" snsnifferEnable	#to use snsniffer for ap sta statistic function
				else
					config_get airodumpEnable "$vif" airodumpEnable	#to use airodump for ap sta statistic function
				fi


				if [ "$airodumpEnable" == 1 && $sniffer_enable -ne 1 ]; then
					ap_sta_file_prefix="ap_sta_list_"
					ap_sta_list="$scanRadioFolder""$ap_sta_file_prefix""$ifname"
					chkAirodump=$(type airodump-ng | grep is)
					if [ -n "$chkAirodump" ]; then	#airodump tool exists
						[ -z "$airodAgeTime" ] && airodAgeTime=$DefaultAgeOutTime
						screen -d -m airodump-ng $ifname -w $ap_sta_list -K 1 --output-format csv -s 1
					else
						DebugPrint "airodump-ng not exists"
					fi
				fi

				if [ "$snsnifferEnable" == 1 && $sniffer_enable -eq 1 ]; then
					ap_sta_file_prefix="ap_sta_list_"
					ap_sta_list="$scanRadioFolder""$ap_sta_file_prefix""$ifname"
					chksnsniffer=$(type snsniffer | grep is)
					if [ -n "$chksnsniffer" ]; then	#snsniffer tool exists
						screen -d -m snsniffer $ifname -w $ap_sta_list 
					else
						DebugPrint "snsniffer not exists"
					fi
				fi				
                        }
                done
        }
}

check_scan_radio_enable() {

	isScanRadioEnable=0
	checkRadio=$1
	wifi_devs=$(uci show |grep opmode=\'mon\' | awk -F '.' '{print $2}')

	for wifi_dev in $wifi_devs; do
		[ "$isScanRadioEnable" == "1" ] && break

		config_get disabled_dev "$wifi_dev" disabled
		[ "$disabled_dev" == "1" ] && continue

		vifs=$(eval "/usr/sbin/foreach wireless wifi-iface device $wifi_dev")

		for vif in $vifs; do
			config_get mode "$vif" mode
			config_get disabled "$vif" disabled

			if [  "$mode" == "monitor" -a "$disabled" == "0" ]; then
				config_get scanProgress "$vif" scanProgress
				
				if [ $sniffer_enable -eq 1 ]; then 
					config_get snsnifferEnable "$vif" snsnifferEnable
					[ "$scanProgress" == "1" -a "$snsnifferEnable" == "1" ] && {
						isScanRadioEnable=1
						break
					}
				else
					config_get airodumpEnable "$vif" airodumpEnable
					[ "$scanProgress" == "1" -a "$airodumpEnable" == "1" ] && {
						isScanRadioEnable=1
						break
					}
				fi
			fi
		done
	done

	echo "$isScanRadioEnable"
}

get_backgroundscan_vif(){

	wifi_dev=$1
	bg_vif=$(ps -w | grep BackgroundScan.sh | grep $wifi_dev | awk '{print $8}')
	echo "$bg_vif"
}

disable_background_scan() {

	wifi_dev=$1

	[ ! -e "/sbin/BackgroundScan.sh" ] && return

	scan_if=$(uci -q get wireless.$wifi_dev"_scan".ifname)

	PID_BGScan=$(ps -w | grep BackgroundScan.sh | grep $wifi_dev | awk '{print $1}')
	
	if [ $sniffer_enable -eq 1 ]; then 
		PID_snsniffer=$(ps -w | grep snsniffer | grep $scan_if | awk '{print $1}')
		kill -9 $PID_BGScan $PID_snsniffer
	else
		PID_Airo=$(ps -w | grep airodump-ng | grep $scan_if | awk '{print $1}')
		kill -9 $PID_BGScan $PID_Airo
	fi

	kill -9 $PID_BGScan $PID_Airo
	rm /tmp/Background_Scanning/$wifi_dev* -f  #remove with determined wifi_dev and scan_if
	rm /tmp/Background_Scanning/$scan_if* -f

	config_get apsteer "$wifi_dev" apsteer 0
	config_get backgroundscanEnable "$wifi_dev" backgroundscanEnable 0
	config_get vifs "$device" vifs

	for vif in $vifs; do
		config_get mode "$vif" mode
		[ "$mode" == "monitor" ] && {
			[ "$apsteer" -eq "0" -a "$backgroundscanEnable" -eq "0" ] && {
				config_get disabled "$vif" disabled
				[ "$disabled" -eq "0" ] && {
					uci -q set wireless.$vif.disabled="1"
					config_set "$vif" disabled 1
					uci commit wireless
				}
			}
		}
	done
}

scan_cpu_adjust() {
    #Senao George, enhanced throuput about 50Mbps in 5G when bat0 use CPU3(4) and scan_1 use CPU4(8)
    #$1 : ifname
    SCANList=`ls /sys/class/net/ | grep "^$1"`

    [ $(get_cpu_num) -eq 4 ] || break

    echo ">>>>> scan_cpu_adjust: SCANList:$SCANList" > /dev/console

    for scan in $SCANList
    do
        for rxqueue in `find /sys/class/net/$scan/queues/ -name "rx-*"`
        do
            echo 8 > $rxqueue/rps_cpus
        done
    done  
}

set_background_scan() {

	local wifi_dev=$1
	local isFastroaming=0 # if 1, use exttool to switch channel	
	local backgroundscan=1
	local fastroamingOnly=2
	local backgroundscan_fastroaming=3
	local apsteerOnly=4
	local backgroundscan_apsteer=5
	local bgmode=0
	local bg_vif
	local ap_vif

	config_get opmode "$wifi_dev" opmode

	if [ "$opmode" == "ap" -o "$opmode" == "wds_ap" ]; then {
		config_get backgroundscanEnable "$wifi_dev" backgroundscanEnable
		config_get vifs "$wifi_dev" vifs
		config_get apsteer_enable "$wifi_dev" apsteer 0

		# senao WAR : mesh enable -> SCAN SSID disable
		[ "$(uci -q get wifiprofile.snWifiConf.SUPPORT_WAR_MESH_DISABLE_SCAN_SSID)" == "1" -a "$(uci get mesh.wifi.disabled)" == "0" ] && {
			backgroundscanEnable=0
			apsteer_enable=0
		}

		# senao WAR : mpsk enable -> SCAN SSID and Background disable
		mpsk_enable=0
		mpsk_ifs=$(foreach wireless wifi-iface multi_group_key 1)
		for mpsk_if in $mpsk_ifs; do
			[ "$(uci get wireless.$mpsk_if.disabled)" = "0" ] && mpsk_enable=1
		done
		[ "$(uci -q get wifiprofile.snWifiConf.SUPPORT_WAR_MPSK_DISABLE_BACKGROUNDSCAN)" == "1" -a $mpsk_enable -eq 1 ] && {
			backgroundscanEnable=0
			apsteer_enable=0
		}


		if [ "$apsteer_enable" == 1 ]; then	#wireless.wifix.apsteer and wifisyncd.wifisyncd.enabled should be enabled to start apsteer function
			wifisyncd_enable=$(uci get wifisyncd.wifisyncd.enabled)
			if [ "$wifisyncd_enable" != "1" ]; then
				apsteer_enable=0
			fi
		fi

		[ -z $vifs ] && vifs=$(eval "/usr/sbin/foreach wireless wifi-iface device $wifi_dev")

# Make 11r(fastroaming) independent from background scan. Not to check 11r to decide whether background scan turns on or not.
#		for vif in $vifs; do
#			config_get ifname "$vif" ifname
#			config_get disabled "$vif" disabled
#			config_get fastroamingEnable "$vif" fastroamingEnable
#
#			if [ "$disabled" == "0" ]; then {
#				config_get mode "$vif" mode
#				[ "$mode" == "ap" ] && [ -z $ap_vif ] && ap_vif=$vif
#				[ "$fastroamingEnable" == "1" ] && {
#					isFastroaming=1
#					break
#				}
#			}
#			fi
#		done

		haveScanRadio=$(check_scan_radio_enable $wifi_dev)

		if [ $sniffer_enable -eq 1 ]; then 
			[ "$backgroundscanEnable" == "1" -o "$apsteer_enable" == "1" ] && [ "$haveScanRadio" != "1" ] && { # vif must be a monitor mode to enable snsniffer
	        		for vif in $vifs; do
					config_get mode "$vif" mode
                			config_get disabled "$vif" disabled
					[ "$mode" == "monitor" -a "$disabled" == "0" ] && {
						bg_vif=$vif
						break
					}
				done
			}
		else
			[ "$backgroundscanEnable" == "1" -o "$apsteer_enable" == "1" ] && [ "$haveScanRadio" != "1" ] && { # vif must be a monitor mode to enable airodump-ng
		        	for vif in $vifs; do
					config_get mode "$vif" mode
        	        		config_get disabled "$vif" disabled
					[ "$mode" == "monitor" -a "$disabled" == "0" ] && {
						bg_vif=$vif
						break
					}
				done
			}
		fi

		if [ -e "/sbin/BackgroundScan.sh" ]; then
			if [ "$backgroundscanEnable" == "1" -a "$apsteer_enable" == "1" ]; then
				bgmode=$backgroundscan_apsteer
			elif [ "$backgroundscanEnable" == "1" -a "$isFastroaming" == "0" ]; then
				bgmode=$backgroundscan
			elif [ "$backgroundscanEnable" == "1" -a "$isFastroaming" == "1" ]; then
				bgmode=$backgroundscan_fastroaming
			elif [ "$backgroundscanEnable" == "0" -a "$apsteer_enable" == "0" -a "$isFastroaming" == "1" ]; then
				bgmode=$fastroamingOnly
			elif [ "$backgroundscanEnable" == "0" -a "$apsteer_enable" == "1" ]; then
				bgmode=$apsteerOnly
			else
				DebugPrint "No backgroundscan mode"
				DebugPrint "backgroundscanEnable=$backgroundscanEnable, isFastroaming=$isFastroaming, apsteer_enable=$apsteer_enable"
			fi

			# senao WAR : mpsk enable -> SCAN SSID and Background disable
			[ "$(uci -q get wifiprofile.snWifiConf.SUPPORT_WAR_MPSK_DISABLE_BACKGROUNDSCAN)" == "1" -a $mpsk_enable -eq 1 ] && bgmode=0

			if [ "$bgmode" != "0" ]; then
				capwapEnabled=$(uci -q get wifiprofile.snWifiConf.SUPPORT_CAPWAP_IWLIST_SCAN)	#for EWS series WTP, not to use monitor mode to scan, just exttool change channel and displayscan
				if [ "$capwapEnabled" == "1" ]; then
					bgmode=$fastroamingOnly	 #we can just use this mode since EWS doesn't have apsteering, rogue ap detection use iwlist displayscan
					bg_vif=$ap_vif
				fi

				if [ -n "$bg_vif" ]; then
					sh /sbin/BackgroundScan.sh $wifi_dev $bg_vif $bgmode $haveScanRadio &
				else
					sh /sbin/BackgroundScan.sh $wifi_dev $vif $bgmode $haveScanRadio &
				fi
			else
				DebugPrint "do nothing for background scanning"
			fi
		else
			DebugPrint "/sbin/BackgroundScan.sh not exists"
		fi

		for vif in $vifs; do
			config_get ifname "$vif" ifname
			config_get mode "$vif" mode
			config_get disabled "$vif" disabled
			[ "$mode" == "monitor" -a "$disabled" == "0" -a "$haveScanRadio" != "1" ] && scan_cpu_adjust $ifname
		done

	}
	fi
}

set_wifi_led() {
	
	wifi_dev=$1
	if [ "$wifi_dev" == "wifi0" ]; then
		led_label="wlan_2g_led"
	elif [ "$wifi_dev" == "wifi1" ]; then
		led_label="wlan_5g_led"
	elif [ "$wifi_dev" == "wifi2" ]; then
		led_label="wlan_5g_2_led"
	else
		DebugPrint "==ERROR===, not support wifi_dev=$wifi_dev"	
		return
	fi		
	sys_class_led_path="/sys/class/leds/$led_label"
	echo "netdev" > "${sys_class_led_path}/trigger"
	echo $wifi_dev > "${sys_class_led_path}/device_name"
	echo "link tx" > "${sys_class_led_path}/mode"
	
}

set_rfdetector() {
	
	wifi_dev=$1

	config_get rfdetect "$wifi_dev" rfdetect 0

	PID_rfdetect=$(ps -w | grep rfdetect | grep $wifi_dev | awk '{print $1}')

	if [ ! -f /sbin/rfdetector ]; then
		echo ">>>>>Install rfdetector first to enable rfdetect" > /dev/console
		return
	fi

	#Delete old process
	if [ -n "$PID_rfdetect" ]; then
		echo ">>>>>PID:$PID_rfdetect Rfdetect is delete" > /dev/console
		kill -9 $PID_rfdetect
	fi

	if [ "$rfdetect" == "1" ]; then
		echo "===Rfdetect is enable in $wifi_dev===" > /dev/console
		rfdetector -i $wifi_dev --out 1 --mode lite &
	else
		echo "===Rfdetect is not enable in $wifi_dev===" > /dev/console
	fi
	
}


set_bridge_mac_to_proc() {
	support_senaolog=$(get_sn_wifi_option SUPPORT_SENAOLOG)

	[ "$support_senaolog" == "0" ] && return

	bridge_mac=$(uci get network.lan.macaddr)

	[ -z "$bridge_mac" ] && bridge_mac=$(ifconfig br-lan | grep HWaddr | awk -F " " '{print $5}')

	echo $bridge_mac > /proc/bridge_mac
}

function update_ini_file()
{
	DebugPrint "update ini $1=$2"
        grep -q $1 /ini/global.ini && sed -i "/$1=/c $1=$2" /ini/global.ini || echo $1=$2 >> /ini/global.ini
}

function update_internal_ini()
{
        grep -q $2 /ini/internal/$1 && sed -i "/$2=/c $2=$3" /ini/internal/$1 || echo $2=$3 >> /ini/internal/$1
        sync
}

function update_ini_for_lowmem()
{
	update_internal_ini $1 dp_rxdma_monitor_buf_ring 128
	update_internal_ini $1 dp_rxdma_monitor_dst_ring 128
	update_internal_ini $1 dp_rxdma_monitor_desc_ring 4096
	update_internal_ini $1 dp_rxdma_monitor_status_ring 512
	sync
}

function update_ini_for_hk_sbs()
{
	update_internal_ini $1 dp_tx_device_limit 49152
	sync
}

function update_ini_for_hk_dbs()
{
	update_internal_ini $1 dp_tx_device_limit 49152
	sync
}

function update_ini_nss_info()
{
	[ -f /lib/wifi/wifi_nss_hk_olnum ] && { \
		local hk_ol_num="$(cat /lib/wifi/wifi_nss_hk_olnum)"
		if [ -e /sys/firmware/devicetree/base/MP_256 ]; then
			update_internal_ini QCA8074V2_i.ini dp_nss_comp_ring_size 0x2000
		elif [ -e /sys/firmware/devicetree/base/MP_512 ]; then
			if [ $hk_ol_num -eq 3 ]; then
				update_internal_ini QCA8074V2_i.ini dp_nss_comp_ring_size 0x3000
			else
				update_internal_ini QCA8074V2_i.ini dp_nss_comp_ring_size 0x4000
			fi
		else
			if [ $hk_ol_num -eq 3 ]; then
				update_internal_ini QCA8074V2_i.ini dp_nss_comp_ring_size 0x8000
			else
				update_internal_ini QCA8074V2_i.ini dp_nss_comp_ring_size 0x8000
			fi
		fi
	}
	sync
}

#function update_ini_for_512MP()
#{
#	update_internal_ini $1 dp_tx_desc 0x4000
#	sync
#}

function update_ini_for_512MP_dp_tx_desc()
{
	update_internal_ini $1 dp_tx_desc 0x4000
	update_internal_ini $1 dp_tx_desc_limit_0 8192
	update_internal_ini $1 dp_tx_desc_limit_1 8192
	update_internal_ini $1 dp_tx_desc_limit_2 8192
	sync
}

function update_ini_for_512MP_P_build()
{
	local board_name

	[ -f /tmp/sysinfo/board_name ] && {
		board_name=$(cat /tmp/sysinfo/board_name)
	}
	update_internal_ini $1 dp_rxdma_monitor_buf_ring 128
	update_internal_ini $1 dp_rxdma_monitor_dst_ring 128
	update_internal_ini $1 dp_rxdma_monitor_desc_ring 4096

	case "$board_name" in
	ap-cp*)
        update_internal_ini $1 dp_rxdma_monitor_status_ring 1024
		;;
	*)
        update_internal_ini $1 dp_rxdma_monitor_status_ring 512
		;;
	esac
	sync
}

function update_ini_for_monitor_buf_ring()
{
	update_internal_ini $1 dp_rxdma_monitor_buf_ring 4096
	update_internal_ini $1 dp_rxdma_monitor_dst_ring 4096
	update_internal_ini $1 dp_rxdma_monitor_desc_ring 4096
	update_internal_ini $1 dp_rxdma_monitor_status_ring 1024
	sync
}


set_global_ini() {

	echo -n "/ini" > /sys/module/firmware_class/parameters/path	#default path for global.ini

	if [ "$default_cfg80211" == "1" ]; then
		update_ini_file cfg80211_config "1"
	else	
		update_ini_file cfg80211_config "0"
	fi
        config_get vow_config qcawifi vow_config
        [ -n "$vow_config" ] && update_ini_file vow_config "$vow_config"

        config_get carrier_vow_config qcawifi carrier_vow_config
        [ -n "$carrier_vow_config" ] && update_ini_file carrier_vow_config "$carrier_vow_config"

        config_get fw_vow_stats_enable qcawifi fw_vow_stats_enable
        [ -n "$fw_vow_stats_enable" ] && update_ini_file fw_vow_stats_enable "$fw_vow_stats_enable"

        config_get ol_bk_min_free qcawifi ol_bk_min_free
        [ -n "$ol_bk_min_free" ] && update_ini_file OL_ACBKMinfree "$ol_bk_min_free"

        config_get ol_be_min_free qcawifi ol_be_min_free
        [ -n "$ol_be_min_free" ] && update_ini_file OL_ACBEMinfree "$ol_be_min_free"

        config_get ol_vi_min_free qcawifi ol_vi_min_free
        [ -n "$ol_vi_min_free" ] && update_ini_file OL_ACVIMinfree "$ol_vi_min_free"

        config_get ol_vo_min_free qcawifi ol_vo_min_free
        [ -n "$ol_vo_min_free" ] && update_ini_file OL_ACVOMinfree "$ol_vo_min_free"

        config_get fw_dump_options qcawifi fw_dump_options
        [ -n "$fw_dump_options" ] && update_ini_file fw_dump_options "$fw_dump_options"

        config_get enableuartprint qcawifi enableuartprint
        [ -n "$enableuartprint" ] && update_ini_file enableuartprint "$enableuartprint"

        config_get max_descs qcawifi max_descs
        [ -n "$max_descs" ] && update_ini_file max_descs "$max_descs"

        config_get max_peers qcawifi max_peers
        [ -n "$max_peers" ] && update_ini_file max_peers "$max_peers"

        config_get cce_disable qcawifi cce_disable
        [ -n "$cce_disable" ] && update_ini_file cce_disable "$cce_disable"

        config_get qwrap_enable qcawifi qwrap_enable
        [ -n "$qwrap_enable" ] && update_ini_file qwrap_enable "$qwrap_enable"

        config_get otp_mod_param qcawifi otp_mod_param
        [ -n "$otp_mod_param" ] && update_ini_file otp_mod_param "$otp_mod_param"

        config_get max_active_peers qcawifi max_active_peers
        [ -n "$max_active_peers" ] && update_ini_file max_active_peers "$max_active_peers"

        config_get enable_smart_antenna qcawifi enable_smart_antenna
        [ -n "$enable_smart_antenna" ] && update_ini_file enable_smart_antenna "$enable_smart_antenna"

        config_get sa_validate_sw qcawifi sa_validate_sw
        [ -n "$sa_validate_sw" ] && update_ini_file sa_validate_sw "$sa_validate_sw"

        config_get nss_wifi_olcfg qcawifi nss_wifi_olcfg
        if [ -n "$nss_wifi_olcfg" ]; then
                update_ini_file nss_wifi_olcfg "$nss_wifi_olcfg"
                config_get nss_wifi_nxthop_cfg qcawifi nss_wifi_nxthop_cfg
                if [ -n "$nss_wifi_nxthop_cfg" ]; then
                    update_ini_file nss_wifi_nxthop_cfg "$nss_wifi_nxthop_cfg"
                fi
        elif [ -f /lib/wifi/wifi_nss_olcfg ]; then
                nss_wifi_olcfg="$(cat /lib/wifi/wifi_nss_olcfg)"

                if [ $nss_wifi_olcfg != 0 ]; then
                        if [ -f /lib/wifi/wifi_nss_override ] && [ $(cat /lib/wifi/wifi_nss_override) = 1 ]; then
                                echo "NSS offload disabled due to unsupported config" >&2
                                update_ini_file nss_wifi_olcfg 0
                        else
                                update_ini_file nss_wifi_olcfg "$nss_wifi_olcfg"
                                update_ini_file rx_hash 0
                        fi
                else
                        update_ini_file nss_wifi_olcfg 0
                fi
        fi

        config_get max_clients qcawifi max_clients
        [ -n "$max_clients" ] && update_ini_file max_clients "$max_clients"

	config_get enable_rdk_stats qcawifi enable_rdk_stats
	[ -n "$enable_rdk_stats" ] && update_ini_file enable_rdk_stats "$enable_rdk_stats"

        config_get max_vaps qcawifi max_vaps
        [ -n "$max_vaps" ] && update_ini_file max_vaps "$max_vaps"

        config_get enable_smart_antenna_da qcawifi enable_smart_antenna_da
        [ -n "$enable_smart_antenna_da" ] && update_ini_file enable_smart_antenna_da "$enable_smart_antenna_da"

        config_get lteu_support qcawifi lteu_support
        [ -n "$lteu_support" ] && update_ini_file lteu_support "$lteu_support"

        config_get tgt_sched_params qcawifi tgt_sched_params
        [ -n "$tgt_sched_params" ] && update_ini_file tgt_sched_params "$tgt_sched_params"

        config_get enable_eapol_minrate qcawifi enable_eapol_minrate
        [ -n "$enable_eapol_minrate" ] && update_ini_file eapol_minrate_set "$enable_eapol_minrate"

        config_get set_eapol_minrate_ac qcawifi set_eapol_minrate_ac
        [ -n "$set_eapol_minrate_ac" ] && update_ini_file eapol_minrate_ac_set "$set_eapol_minrate_ac"

        config_get enable_mesh_support qcawifi enable_mesh_support
        [ -n "$enable_mesh_support" ] && update_ini_file mesh_support "$enable_mesh_support"

        config_get beacon_offload_disable qcawifi beacon_offload_disable
        [ -n "$beacon_offload_disable" ] && update_ini_file beacon_offload_disable "$beacon_offload_disable"

	#jacky: removed in spf10.0_fc
	#config_get hw_mode_id qcawifi hw_mode_id
	#[ -n "$hw_mode_id" ] && update_ini_file hw_mode_id "$hw_mode_id"

	config_get spectral_disable qcawifi spectral_disable
	[ -n "$spectral_disable" ] && update_ini_file spectral_disable "$spectral_disable"

        config_get twt_enable qcawifi twt_enable
        [ -n "$twt_enable" ] && update_ini_file twt_enable "$twt_enable"

        config_get mbss_ie_enable qcawifi mbss_ie_enable
        [ -n "$mbss_ie_enable" ] && update_ini_file mbss_ie_enable "$mbss_ie_enable"

}

atf_config() {
	config_load wireless
	config_foreach atf_radio_vap_params_config wifi-device

	config_load wireless
	config_foreach atf_group_config atf-config-group

	config_load wireless
	config_foreach atf_ssid_config atf-config-ssid

	config_load wireless
	config_foreach atf_sta_config atf-config-sta

	config_load wireless
	config_foreach atf_ac_config atf-config-ac

	config_load wireless
	config_foreach atf_tput_config atf-config-tput

	config_load wireless
	config_foreach atf_enable wifi-device
}

atf_radio_vap_params_config() {
	local device="$1"
	local atf_sched_dur
	local atfstrictsched
	local atfobsssched
	local atfobssscale
	local atfgrouppolicy

	config_get atf_sched_dur "$device" atf_sched_dur
	[ -n "$atf_sched_dur" ] && device_if "$device" "atf_sched_dur" "$atf_sched_dur"

	config_get atfstrictsched "$device" atfstrictsched
	[ -n "$atfstrictsched" ] && device_if "$device" "atfstrictsched" "$atfstrictsched"

	config_get atfobsssched "$device" atfobsssched
	[ -n "$atfobsssched" ] && device_if "$device" "atfobsssched" "$atfobsssched"

	config_get atfobssscale "$device" atfobssscale
	[ -n "$atfobssscale" ] && device_if "$device" "atfobssscale" "$atfobssscale"

	config_get atfgrouppolicy "$device" atfgrouppolicy
	[ -n "$atfgrouppolicy" ] && device_if "$device" "atfgrouppolicy" "$atfgrouppolicy"

	config_get disabled $device disabled 0
	if [ $disabled -eq 0 ]; then
		config_get vifs "$device" vifs

		local ifidx=0
		local radioidx=${device#wifi}
		for vif in $vifs; do
			local vifname
			[ $ifidx -gt 0 ] && vifname="ath${radioidx}$ifidx" || vifname="ath${radioidx}"

			config_get atf_shr_buf "$vif" atf_shr_buf
			[ -n "$atf_shr_buf" ] && device_if "$vifname" "atf_shr_buf" "$atf_shr_buf"

			config_get atf_max_buf "$vif" atf_max_buf
			[ -n "$atf_max_buf" ] && device_if "$vifname" "atf_max_buf" "$atf_max_buf"

			config_get atf_min_buf "$vif" atf_min_buf
			[ -n "$atf_min_buf" ] && device_if "$vifname" "atf_min_buf" "$atf_min_buf"

			config_get commitatf "$vif" commitatf
			[ -n "$commitatf" ] && device_if "$vifname" "commitatf" "$commitatf"

			config_get atfmaxclient "$vif" atfmaxclient
			[ -n "$atfmaxclient" ] && device_if "$vifname" "atfmaxclient" "$atfmaxclient"

			config_get atfssidgroup "$vif" atfssidgroup
			[ -n "$atfssidgroup" ] && device_if "$vifname" "atfssidgroup" "$atfssidgroup"

			config_get atf_tput_at "$vif" atf_tput_at
			[ -n "$atf_tput_at" ] && device_if "$vifname" "atf_tput_at" "$atf_tput_at"

			config_get atfssidsched "$vif" atfssidsched
			[ -n "$atfssidsched" ] && device_if "$vifname" "atfssidsched" "$atfssidsched"

			ifidx=$(($ifidx + 1))
		done
	fi

}

atf_group_config() {
	local cmd
	local group
	local ssid
	local airtime
	local device

	config_get device "$1" device
	radioidx=${device#wifi}

	config_get cmd "$1" command
	config_get group "$1" group
	config_get ssid "$1" ssid
	config_get airtime "$1" airtime

	if [ -z "$cmd" ] || [ -z "$group" ] ; then
		echo "Invalid ATF GROUP Configuration"
		return 1
	fi

	if [ "$cmd" == "addgroup" ] && [ -n "$ssid" ] && [ -n "$airtime" ]; then
		for word in $ssid; do
			if [ "$default_cfg80211" == "1" ]; then
	                        wlanconfig ath$radioidx addatfgroup $group $word -cfg80211
			else
				wlanconfig ath$radioidx addatfgroup $group $word
			fi
		done
		if [ "$default_cfg80211" == "1" ]; then
			wlanconfig ath$radioidx configatfgroup $group $airtime -cfg80211
		else
			wlanconfig ath$radioidx configatfgroup $group $airtime
		fi
	fi

	if [ "$cmd" == "delgroup" ]; then
		if [ "$default_cfg80211" == "1" ]; then
			wlanconfig ath$radioidx delatfgroup $group -cfg80211
		else
			wlanconfig ath$radioidx delatfgroup $group
		fi
	fi
}

atf_ssid_config() {
	local cmd
	local ssid
	local airtime
	local device

	config_get device "$1" device
	radioidx=${device#wifi}

	config_get cmd "$1" command
	config_get ssid "$1" ssid
	config_get airtime "$1" airtime

	if [ -z "$cmd" ] || [ -z "$ssid" ] ; then
		echo "Invalid ATF SSID Configuration"
		return 1
	fi

	if [ "$cmd" == "addssid" ] && [ -n "$airtime" ]; then
		if [ "$default_cfg80211" == "1" ]; then
			wlanconfig ath$radioidx $cmd $ssid $airtime -cfg80211
		else
			wlanconfig ath$radioidx $cmd $ssid $airtime
		fi
	fi

	if [ "$cmd" == "delssid" ]; then
		if [ "$default_cfg80211" == "1" ]; then
			wlanconfig ath$radioidx $cmd $ssid -cfg80211
		else
			wlanconfig ath$radioidx $cmd $ssid
		fi
	fi
}

atf_sta_config() {
	local cmd
	local ssid
	local airtime
	local device
	local mac

	config_get device "$1" device
	radioidx=${device#wifi}

	config_get cmd "$1" command
	config_get airtime "$1" airtime
	config_get ssid "$1" ssid
	config_get mac "$1" macaddr
	mac="${mac//:}"

	if [ -z "$cmd" ] || [ -z "$mac" ] ; then
		echo "Invalid ATF STA Configuration"
		return 1
	fi

	if [ "$cmd" == "addsta" ] && [ -n "$airtime" ]; then
		if [ "$default_cfg80211" == "1" ]; then
			wlanconfig ath$radioidx $cmd $mac $airtime $ssid -cfg80211
		else
			wlanconfig ath$radioidx $cmd $mac $airtime $ssid
		fi
	fi

	if [ "$cmd" == "delsta" ]; then
		if [ "$default_cfg80211" == "1" ]; then
			wlanconfig ath$radioidx $cmd $mac -cfg80211
		else
			wlanconfig ath$radioidx $cmd $mac
		fi
	fi
}

atf_ac_config() {
	local cmd
	local ssid
	local device
	local ac
	local airtime

	config_get device "$1" device
	radioidx=${device#wifi}

	config_get cmd "$1" command
	config_get ac "$1" ac
	config_get airtime "$1" airtime
	config_get ssid "$1" ssid

	if [ -z "$cmd" ] || [ -z "$ssid" ] || [ -z "$ac" ] ; then
		echo "Invalid ATF AC Configuration"
		return 1
	fi

	if [ "$cmd" == "atfaddac" ] && [ -n "$airtime" ]; then
		if [ "$default_cfg80211" == "1" ]; then
			echo "wlanconfig ath$radioidx $cmd $ssid $ac:$airtime" -cfg80211
		else
			echo "wlanconfig ath$radioidx $cmd $ssid $ac:$airtime"
		fi
	fi

	if [ "$cmd" == "atfdelac" ]; then
		if [ "$default_cfg80211" == "1" ]; then
			echo "wlanconfig ath$radioidx $cmd $ssid $ac" -cfg80211
		else
			echo "wlanconfig ath$radioidx $cmd $ssid $ac"
		fi
	fi
}

atf_tput_config() {
	local cmd
	local tput
	local max_airtime
	local device
	local mac

	config_get device "$1" device
	radioidx=${device#wifi}

	config_get cmd "$1" command
	config_get tput "$1" throughput
	config_get max_airtime "$1" max_airtime
	config_get mac "$1" macaddr
	mac="${mac//:}"

	if [ -z "$cmd" ] || [ -z "$mac" ] || [ -z "$tput" ] ; then
		echo "Invalid ATF Throughput Configuration"
		return 1
	fi

	if [ "$cmd" == "addtputsta" ]; then
		device_if ath$radioidx commitatf 0
		if [ "$default_cfg80211" == "1" ]; then
			wlanconfig ath$radioidx addtputsta $mac $tput $max_airtime -cfg80211
		else
			wlanconfig ath$radioidx addtputsta $mac $tput $max_airtime
		fi
	fi

	if [ "$cmd" == "deltputsta" ]; then
		device_if ath$radioidx commitatf 0
		if [ "$default_cfg80211" == "1" ]; then
			wlanconfig ath$radioidx deltputsta $mac -cfg80211
		else
			wlanconfig ath$radioidx deltputsta $mac
		fi
	fi
}

atf_enable() {
	local device="$1"

	config_get disabled $device disabled 0
	if [ $disabled -eq 0 ]; then
		config_get vifs "$device" vifs
		echo "device: $device vifs: $vifs"

		local ifidx=0
		local radioidx=${device#wifi}
		for vif in $vifs; do
			local vifname
			[ $ifidx -gt 0 ] && vifname="ath${radioidx}$ifidx" || vifname="ath${radioidx}"

			config_get commitatf "$vif" commitatf
			[ -n "$commitatf" ] && device_if "$vifname" "commitatf" "$commitatf"

			ifidx=$(($ifidx + 1))
		done
	fi
}

notify_ubus_to_unbridge()
{
	local interface=$1
	local tmp=$(get_bridge_name $interface)
	local network=${tmp#br-}
	local counter=0

	if [ $interface == "ath32" -o $interface == "ath35" ]; then
		return	
	fi
	while [ -z "$network" -a $counter -lt 5 ]
	do
		#echo -------------- > /dev/console
		#brctl show  > /dev/console
		tmp=$(get_bridge_name $1)
		network=${tmp#br-}
		sleep 1
		counter=$(($counter + 1))
	done

        #echo ==========interface=$interface network=$network =========== > /dev/console
	if [ -n "$network" ]; then
		ubus call network.interface."$network" remove_device "{ \"name\" : \"$interface\"}"
	else
		DebugPrint "$interface not in bridge, no need to notify ubus to unbridge"
	fi

	return

}

set_legacy_deny()
{
    local ignore_legacy=0
    local wifi_dev=$1

    config_get opmode "$wifi_dev" opmode
    config_get legacy_hwmode_deny "$wifi_dev" legacy_hwmode_deny

    if [ "$legacy_hwmode_deny" != "0" -a "$legacy_hwmode_deny" != "1" ]; then
        return
    fi

    if [ "$opmode" == "ap" -o "$opmode" == "wds_ap" ]; then
        config_get vifs "$wifi_dev" vifs

	[ -z $vifs ] && vifs=$(eval "/usr/sbin/foreach wireless wifi-iface device $wifi_dev")

        for vif in $vifs; do
            config_get ifname "$vif" ifname
            config_get disabled "$vif" disabled
            config_get mode_display "$vif" mode_display

            [ "$disabled" == "0" -a "$mode_display" == "ap" ] && {
                device_if "$ifname" puren "$legacy_hwmode_deny"
                #echo "interface:[$vif] legacy_hwmode_deny:[$legacy_hwmode_deny]" > /dev/console
            }
        done
    fi
}

mesh_RP_config()
{
	if [ "$(uci get mesh.wifi.disabled)" == "0" ]; then
		[ -e "/tmp/myNextHopMAC" ] && rm /tmp/myNextHopMAC
		mac_str=$(cat /tmp/wifi0_mac.txt | sed 's/://g')
		uci set wireless.wifi0_rp_ap.ssid="MESH_RP_AP${mac_str:6:6}"
		mac_str=$(cat /tmp/wifi1_mac.txt | sed 's/://g')
		uci set wireless.wifi1_rp_ap.ssid="MESH_RP_AP${mac_str:6:6}"
		mesh_key=$(uci get wireless.wifi0_mesh.aeskey)
		uci set wireless.wifi0_rp_ap.key=$mesh_key
		uci set wireless.wifi0_rp_sta.key=$mesh_key
		uci set wireless.wifi1_rp_ap.key=$mesh_key
		uci set wireless.wifi1_rp_sta.key=$mesh_key

		if [ "$(uci -q get wireless.wifi0_mesh.disabled)" == "0" ]; then
			uci set wireless.wifi0_rp_ap.disabled=0
			uci set wireless.wifi1_rp_ap.disabled=1
			uci set wireless.wifi0_rp_sta.disabled=0
			uci set wireless.wifi1_rp_sta.disabled=1
		elif [ "$(uci -q get wireless.wifi1_mesh.disabled)" == "0" ]; then
			uci set wireless.wifi0_rp_ap.disabled=1
			uci set wireless.wifi1_rp_ap.disabled=0
			uci set wireless.wifi0_rp_sta.disabled=1
			uci set wireless.wifi1_rp_sta.disabled=0
		fi
	fi
}

ftm_qcawifi() {
	error=0
	local board_name
	[ -f /tmp/sysinfo/board_name ] && {
		board_name=$(cat /tmp/sysinfo/board_name)
	}

	echo -n "/ini" > /sys/module/firmware_class/parameters/path
	case "$board_name" in
	ap-hk*|ap-dk*|ap-ac*|ap-cp*|ap-oa*)
		echo "Entering FTM mode operation" > /dev/console
	;;
	*)
		echo "FTM mode operation not applicable. Returning" > /dev/console
		return
	;;
	esac

	rm -rf /etc/config/wireless

	for mod in $(cat /lib/wifi/qca-wifi-modules); do

		case ${mod} in
			umac) [ -d /sys/module/${mod} ] || { \
				insmod ${mod} || { \
					lock -u /var/run/wifilock
					unload_qcawifi
					error=1
				}
			};;

			qdf) [ -d /sys/module/${mod} ] || { \
				insmod ${mod} || { \
					lock -u /var/run/wifilock
					unload_qcawifi
					error=2
				}
			};;

			qca_ol) [ -d /sys/module/${mod} ] || { \
				do_cold_boot_calibration
				insmod ${mod} testmode=1 hw_mode_id=1 cfg80211_config=1 || { \
					lock -u /var/run/wifilock
					unload_qcawifi
					error=3
				}
			};;

			qca_da|ath_dev|hst_tx99|ath_rate_atheros|ath_hal)
			;;

			smart_antenna|ath_pktlog)
			;;

			*) [ -d /sys/module/${mod} ] || { \
				insmod ${mod} || { \
					lock -u /var/run/wifilock
					unload_qcawifi
					error=4
				}
			};;

		esac
	done

	case "$board_name" in
	ap-hk*|ap-ac*|ap-cp*|ap-oa*)
		rm -rf /etc/modules.d
		mv /etc/modules.d.bk /etc/modules.d
	;;
	*)
	;;
	esac
	sync
	[ $error != 0 ] && echo "FTM error: $error" > /dev/console && return 1
	dmesg -n8
	ftm -n &
	#dmesg got disabled earlier in boot-ftm file
	#enable dmesg back
	echo "FTM mode interface is ready now" > /dev/kmsg
}

waltest_qcawifi() {
	error=0
	local board_name
	[ -f /tmp/sysinfo/board_name ] && {
		board_name=$(cat /tmp/sysinfo/board_name)
	}

	echo -n "/ini" > /sys/module/firmware_class/parameters/path
	case "$board_name" in
	ap-hk*)
		echo "Entering WALtest mode operation" > /dev/console
	;;
	*)
		echo "WALtest mode operation not applicable. Returning" > /dev/console
		return
	;;
	esac

	rm -rf /etc/config/wireless

	for mod in $(cat /lib/wifi/qca-wifi-modules); do

		case ${mod} in
			umac) [ -d /sys/module/${mod} ] || { \
				insmod ${mod} || { \
					lock -u /var/run/wifilock
					unload_qcawifi
					error=1
				}
			};;

			qdf) [ -d /sys/module/${mod} ] || { \
				insmod ${mod} || { \
					lock -u /var/run/wifilock
					unload_qcawifi
					error=2
				}
			};;

			qca_ol) [ -d /sys/module/${mod} ] || { \
				insmod ${mod} testmode=3 hw_mode_id=1 cfg80211_config=1 || { \
					lock -u /var/run/wifilock
					unload_qcawifi
					error=3
				}
			};;

			qca_da|ath_dev|hst_tx99|ath_rate_atheros|ath_hal)
			;;

			smart_antenna|ath_pktlog)
			;;

			*) [ -d /sys/module/${mod} ] || { \
				insmod ${mod} || { \
					lock -u /var/run/wifilock
					unload_qcawifi
					error=4
				}
			};;

		esac
	done

	[ $error != 0 ] && echo "WALtest error: $error" > /dev/console && return 1
	echo "WALtest mode interface is ready now" > /dev/kmsg
}

set_vap_wmm_dscp()
{
	ifname=$1
	device_if $ifname set_dscp_ovride 1
	device_if $ifname s_dscp_tid_map 0xe0 0 #56
	device_if $ifname s_dscp_tid_map 0xc0 0 #48
	device_if $ifname s_dscp_tid_map 0xb8 6 #46
	device_if $ifname s_dscp_tid_map 0xa0 4 #40
	device_if $ifname s_dscp_tid_map 0x98 4 #38
	device_if $ifname s_dscp_tid_map 0x90 4 #36
	device_if $ifname s_dscp_tid_map 0x88 4 #34
	device_if $ifname s_dscp_tid_map 0x80 5 #32
	device_if $ifname s_dscp_tid_map 0x78 4 #30
	device_if $ifname s_dscp_tid_map 0x70 4 #28
	device_if $ifname s_dscp_tid_map 0x68 4 #26
	device_if $ifname s_dscp_tid_map 0x60 4 #24
	device_if $ifname s_dscp_tid_map 0x58 3 #22
	device_if $ifname s_dscp_tid_map 0x50 3 #20
	device_if $ifname s_dscp_tid_map 0x48 3 #18
	device_if $ifname s_dscp_tid_map 0x40 0 #16
	device_if $ifname s_dscp_tid_map 0x38 2 #14
	device_if $ifname s_dscp_tid_map 0x30 2 #12
	device_if $ifname s_dscp_tid_map 0x28 2 #10
	device_if $ifname s_dscp_tid_map 0x20 1 #8
	device_if $ifname s_dscp_tid_map 0x0  0 #0
}

set_wmm_dscp()
{
#
# WMM         802.1p    DSCP
# Voice        7
#              6         46
# Video        5         32
#              4         24, 26, 28, 30, 34, 36, 38, 40
# Best Effort  3         18, 20, 22
#              0         0, 16,48,56
# Background   2         10, 12, 14
#              1         8

	for vap in $(iwconfig 2>/dev/null | grep ^ath | awk {'print $1'}); do
		for iface in $(foreach wireless wifi-iface); do
			if [ "$vap" == "$(uci get wireless.$iface.ifname)" ]; then
				if [ "$(uci get wireless.$iface.mode)" == "ap" ]; then
					device_if $vap set_dscp_ovride 1
					device_if $vap dscp_all 1
				fi
			fi
		done
	done
}

fix_fastroam()
{
#2020.11.26 Senao Sam(written by Senao George), fix fast-roaming issue in IPQ40xx.
#The client will be not connect to AP after several times of roaming.

	for vap in $(iwconfig 2>/dev/null | grep ^ath | awk {'print $1'}); do
		for iface in $(foreach wireless wifi-iface); do
			if [ "$vap" == "$(uci get wireless.$iface.ifname)" ]; then
				if [ "$(uci get wireless.$iface.mode)" == "ap" ] && [ "`cat /tmp/sysinfo/model | grep IPQ40xx`" ] ; then
					wifitool $vap beeliner_fw_test 174 0
				fi
			fi
		done
	done
}

get_cpu_num() {
        local cpunum=`cat /proc/cpuinfo |grep ^processor |wc -l`
    echo ${cpunum:-1}
}

get_cpu_mask() {
    local cpunum=$(get_cpu_num)
        local cpumask=0
        while [ $cpunum -gt 1 ]
        do
                cpunum=$(($cpunum-1))
                cpumask=$(($cpumask+(1 << $cpunum)))
                # remain cpu0 for system use
                [ $cpunum -le 0 ] && break;
        done
        printf %x $cpumask
}

set_throughput_rule() {
        if [ $(get_cpu_num) -gt 1 -a -f /proc/sys/net/core/rps_sock_flow_entries ]
        then
                local cpumask=$(get_cpu_mask)
                local rps_flowentries=`cat /proc/sys/net/core/rps_sock_flow_entries`
                if [ "$rps_flowentries" = "0" ]
                then
                        rps_flowentries=65536
                        echo $rps_flowentries > /proc/sys/net/core/rps_sock_flow_entries
                fi

                if [ $(get_cpu_num) -eq 2 ]
                then
                        VAPList_24G=`ls /sys/class/net/ | grep "^ath0"`
                        for vif in $VAPList_24G
                        do
                                local rps_queue=`find /sys/class/net/$vif/queues/ -name "rx-*"`
                                local rps_queuenum=`echo "$rpsqueue"|wc -l`
                                local rps_flowcnt=`echo $(($rps_flowentries/${rps_queuenum:-1}))`
                                for rxqueue in `find /sys/class/net/$vif/queues/ -name "rx-*"`
                                do
                                        echo 1 > $rxqueue/rps_cpus
                                        echo $rps_flowcnt > $rxqueue/rps_flow_cnt
                                done
                        done
                        VAPList_5G=`ls /sys/class/net/ | grep "^ath1"`
                        for vif in $VAPList_5G
                        do
                                local rps_queue=`find /sys/class/net/$vif/queues/ -name "rx-*"`
                                local rps_queuenum=`echo "$rpsqueue"|wc -l`
                                local rps_flowcnt=`echo $(($rps_flowentries/${rps_queuenum:-1}))`
                                for rxqueue in `find /sys/class/net/$vif/queues/ -name "rx-*"`
                                do
                                        echo 2 > $rxqueue/rps_cpus
                                        echo $rps_flowcnt > $rxqueue/rps_flow_cnt
                                done
                        done
                else
                        VAPList=`ls /sys/class/net/ | grep "^ath"`
                        for vif in $VAPList
                        do
                                local rps_queue=`find /sys/class/net/$vif/queues/ -name "rx-*"`
                                local rps_queuenum=`echo "$rpsqueue"|wc -l`
                                local rps_flowcnt=`echo $(($rps_flowentries/${rps_queuenum:-1}))`
                                for rxqueue in `find /sys/class/net/$vif/queues/ -name "rx-*"`
                                do
                                        echo $cpumask > $rxqueue/rps_cpus
                                        echo $rps_flowcnt > $rxqueue/rps_flow_cnt
                                done
                        done
                fi
        fi

}
